"""
Copyright (c) 2016-2020 The scikit-optimize developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Lhs functions are inspired by
https://github.com/clicumu/pyDOE2/blob/
master/pyDOE2/doe_lhs.py
"""
import numpy as np
from sklearn.utils import check_random_state
from scipy import spatial
from entmoot.sampler.base import InitialPointGenerator
from entmoot.space.space import Space, Categorical



def _random_permute_matrix(h, random_state=None):
    rng = check_random_state(random_state)
    h_rand_perm = np.zeros_like(h)
    samples, n = h.shape
    for j in range(n):
        order = rng.permutation(range(samples))
        h_rand_perm[:, j] = h[order, j]
    return h_rand_perm


class Lhs(InitialPointGenerator):
    """Latin hypercube sampling

    Parameters
    ----------
    lhs_type : str, default='classic'
        - 'classic' - a small random number is added
        - 'centered' - points are set uniformly in each interval

    criterion : str or None, default='maximin'
        When set to None, the LHS is not optimized

        - 'correlation' : optimized LHS by minimizing the correlation
        - 'maximin' : optimized LHS by maximizing the minimal pdist
        - 'ratio' : optimized LHS by minimizing the ratio
          `max(pdist) / min(pdist)`

    iterations : int
        Defines the number of iterations for optimizing LHS
    """
    def __init__(self, lhs_type="classic", criterion="maximin",
                 iterations=1000):
        self.lhs_type = lhs_type
        self.criterion = criterion
        self.iterations = iterations

    def generate(self, dimensions, n_samples, random_state=None):
        """Creates latin hypercube samples.

        Parameters
        ----------
        dimensions : list, shape (n_dims,)
            List of search space dimensions.
            Each search dimension can be defined either as

            - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
              dimensions),
            - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
              dimensions),
            - as a list of categories (for `Categorical` dimensions), or
            - an instance of a `Dimension` object (`Real`, `Integer` or
              `Categorical`).

        n_samples : int
            The order of the LHS sequence. Defines the number of samples.
        random_state : int, RandomState instance, or None (default)
            Set random state to something other than None for reproducible
            results.

        Returns
        -------
        np.array, shape=(n_dim, n_samples)
            LHS set
        """
        rng = check_random_state(random_state)
        space = Space(dimensions)
        transformer = space.get_transformer()
        n_dim = space.n_dims
        space.set_transformer("normalize")
        if self.criterion is None or n_samples == 1:
            h = self._lhs_normalized(n_dim, n_samples, rng)
            h = space.inverse_transform(h)
            space.set_transformer(transformer)
            return h
        else:
            h_opt = self._lhs_normalized(n_dim, n_samples, rng)
            h_opt = space.inverse_transform(h_opt)
            if self.criterion == "correlation":
                mincorr = np.inf
                for i in range(self.iterations):
                    # Generate a random LHS
                    h = self._lhs_normalized(n_dim, n_samples, rng)
                    r = np.corrcoef(np.array(h).T)
                    if len(np.abs(r[r != 1])) > 0 and \
                            np.max(np.abs(r[r != 1])) < mincorr:
                        mincorr = np.max(np.abs(r - np.eye(r.shape[0])))
                        h_opt = h.copy()
                        h_opt = space.inverse_transform(h_opt)
            elif self.criterion == "maximin":
                maxdist = 0
                # Maximize the minimum distance between points
                for i in range(self.iterations):
                    h = self._lhs_normalized(n_dim, n_samples, rng)
                    d = spatial.distance.pdist(np.array(h), 'euclidean')
                    if maxdist < np.min(d):
                        maxdist = np.min(d)
                        h_opt = h.copy()
                        h_opt = space.inverse_transform(h_opt)
            elif self.criterion == "ratio":
                minratio = np.inf

                # Maximize the minimum distance between points
                for i in range(self.iterations):
                    h = self._lhs_normalized(n_dim, n_samples, rng)
                    p = spatial.distance.pdist(np.array(h), 'euclidean')
                    if np.min(p) == 0:
                        ratio = np.max(p) / 1e-8
                    else:
                        ratio = np.max(p) / np.min(p)
                    if minratio > ratio:
                        minratio = ratio
                        h_opt = h.copy()
                        h_opt = space.inverse_transform(h_opt)
            else:
                raise ValueError("Wrong criterion."
                                 "Got {}".format(self.criterion))
            space.set_transformer(transformer)
            return h_opt

    def _lhs_normalized(self, n_dim, n_samples, random_state):
        rng = check_random_state(random_state)
        x = np.linspace(0, 1, n_samples + 1)
        u = rng.rand(n_samples, n_dim)
        h = np.zeros_like(u)
        if self.lhs_type == "centered":
            for j in range(n_dim):
                h[:, j] = np.diff(x) / 2.0 + x[:n_samples]
        elif self.lhs_type == "classic":
            for j in range(n_dim):
                h[:, j] = u[:, j] * np.diff(x) + x[:n_samples]
        else:
            raise ValueError("Wrong lhs_type. Got ".format(self.lhs_type))
        return _random_permute_matrix(h, random_state=rng)



"""
Copyright (c) 2016-2020 The scikit-optimize developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Inspired by https://github.com/jonathf/chaospy/blob/master/chaospy/
distributions/sampler/sequences/halton.py
"""
import numpy as np
from entmoot.sampler.base import InitialPointGenerator
from entmoot.space.space import Space
from sklearn.utils import check_random_state


class Halton(InitialPointGenerator):
    """Creates `Halton` sequence samples.
    In statistics, Halton sequences are sequences used to generate
    points in space for numerical methods such as Monte Carlo simulations.
    Although these sequences are deterministic, they are of low discrepancy,
    that is, appear to be random
    for many purposes. They were first introduced in 1960 and are an example
    of a quasi-random number sequence. They generalise the one-dimensional
    van der Corput sequences.

    For ``dim == 1`` the sequence falls back to Van Der Corput sequence.

    Parameters
    ----------
    min_skip : int
        minimum skipped seed number. When `min_skip != max_skip`
        a random number is picked.
    max_skip : int
        maximum skipped seed number. When `min_skip != max_skip`
        a random number is picked.
    primes : tuple, default=None
        The (non-)prime base to calculate values along each axis. If
        empty or None, growing prime values starting from 2 will be used.
    """
    def __init__(self, min_skip=-1, max_skip=-1, primes=None):
        self.primes = primes
        self.min_skip = min_skip
        self.max_skip = max_skip

    def generate(self, dimensions, n_samples, random_state=None):
        """Creates samples from Halton set.

        Parameters
        ----------
        dimensions : list, shape (n_dims,)
            List of search space dimensions.
            Each search dimension can be defined either as

            - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
              dimensions),
            - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
              dimensions),
            - as a list of categories (for `Categorical` dimensions), or
            - an instance of a `Dimension` object (`Real`, `Integer` or
              `Categorical`).
        n_samples : int
            The order of the Halton sequence. Defines the number of samples.
        random_state : int, RandomState instance, or None (default)
            Set random state to something other than None for reproducible
            results.

        Returns
        -------
        np.array, shape=(n_dim, n_samples)
            Halton set
        """
        rng = check_random_state(random_state)
        if self.primes is None:
            primes = []
        else:
            primes = list(self.primes)
        space = Space(dimensions)
        n_dim = space.n_dims
        transformer = space.get_transformer()
        space.set_transformer("normalize")
        if len(primes) < n_dim:
            prime_order = 10 * n_dim
            while len(primes) < n_dim:
                primes = _create_primes(prime_order)
                prime_order *= 2

            primes = primes[:n_dim]
        assert len(primes) == n_dim, "not enough primes"
        if self.min_skip < 0 and self.max_skip < 0:
            skip = max(primes)
        elif self.min_skip == self.max_skip:
            skip = self.min_skip
        elif self.min_skip < 0 or self.max_skip < 0:
            skip = np.max(self.min_skip, self.max_skip)
        else:
            skip = rng.randint(self.min_skip, self.max_skip)

        out = np.empty((n_dim, n_samples))
        indices = [idx + skip for idx in range(n_samples)]
        for dim_ in range(n_dim):
            out[dim_] = _van_der_corput_samples(
                indices, number_base=primes[dim_])
        out = space.inverse_transform(np.transpose(out))
        space.set_transformer(transformer)
        return out


def _van_der_corput_samples(idx, number_base=2):
    """
    Create `Van Der Corput` low discrepancy sequence samples.

    A van der Corput sequence is an example of the simplest one-dimensional
    low-discrepancy sequence over the unit interval; it was first described in
    1935 by the Dutch mathematician J. G. van der Corput. It is constructed by
    reversing the base-n representation of the sequence of natural numbers
    (1, 2, 3, ...).

    In practice, use Halton sequence instead of Van Der Corput, as it is the
    same, but generalized to work in multiple dimensions.

    Parameters
    ----------
    idx (int, numpy.ndarray):
        The index of the sequence. If array is provided, all values in
        array is returned.
    number_base : int
        The numerical base from where to create the samples from.

    Returns
    -------
    float, numpy.ndarray
        Van der Corput samples.
    """
    assert number_base > 1

    idx = np.asarray(idx).flatten() + 1
    out = np.zeros(len(idx), dtype=float)

    base = float(number_base)
    active = np.ones(len(idx), dtype=bool)
    while np.any(active):
        out[active] += (idx[active] % number_base)/base
        idx //= number_base
        base *= number_base
        active = idx > 0
    return out


def _create_primes(threshold):
    """
    Generate prime values using sieve of Eratosthenes method.

    Parameters
    ----------
    threshold : int
        The upper bound for the size of the prime values.

    Returns
    ------
    List
        All primes from 2 and up to ``threshold``.
    """
    if threshold == 2:
        return [2]

    elif threshold < 2:
        return []

    numbers = list(range(3, threshold+1, 2))
    root_of_threshold = threshold ** 0.5
    half = int((threshold+1)/2-1)
    idx = 0
    counter = 3
    while counter <= root_of_threshold:
        if numbers[idx]:
            idy = int((counter*counter-3)/2)
            numbers[idy] = 0
            while idy < half:
                numbers[idy] = 0
                idy += counter
        idx += 1
        counter = 2*idx+3
    return [2] + [number for number in numbers if number]



import numpy as np

from sklearn.base import clone
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.utils import check_random_state

import sys


class EntingRegressor(BaseEstimator, RegressorMixin):
    """Predict with LightGBM tree model and include model uncertainty 
    defined by distance-based standard estimator.


    Parameters
    ----------
    base_estimator : LGBMRegressor instance, EntingRegressor instance or 
        None (default). If LGBMRegressor instance of None is given: new 
        EntingRegressor is defined. If EntingRegressor is given: base_estimator
        and std_estimator are given to new instance.
    std_estimator : DistanceBasedStd instance,
        Determines which measure is used to capture uncertainty.
    random_state : int, RandomState instance, or None (default)
        Set random state to something other than None for reproducible
        results.
    """

    def __init__(self, base_estimator=None,
                std_estimator=None,
                random_state=None):
        self.random_state = random_state
        self.base_estimator = base_estimator
        self.std_estimator = std_estimator

        # check if base_estimator is EntingRegressor
        if isinstance(base_estimator, EntingRegressor):
            self.base_estimator = base_estimator.base_estimator
            self.std_estimator = base_estimator.std_estimator

    def fit(self, X, y):
        """Fit model and standard estimator to observations.

        Parameters
        ----------
        X : array-like, shape=(n_samples, n_features)
            Training vectors, where `n_samples` is the number of samples
            and `n_features` is the number of features.

        y : array-like, shape=(n_samples,)
            Target values (real numbers in regression)

        Returns
        -------
        -
        """
        base_estimator = self.base_estimator

        # suppress lgbm output
        base_estimator.set_params(
            random_state=self.random_state,
            verbose=-1
        )

        # clone base_estimator (only supported for sklearn estimators)
        self.regressor_ = clone(base_estimator)

        # update std estimator
        self.std_estimator.update(X, y)

        # update tree model regressor
        self.regressor_.fit(X, y)

    def set_params(self, **params):
        """Sets parameters related to tree model estimator. All parameter 
        options for LightGBM are given here: 
            https://lightgbm.readthedocs.io/en/latest/Parameters.html

        Parameters
        ----------
        kwargs : dict
            Additional arguments to be passed to the tree model

        Returns
        -------
        -
        """
        self.base_estimator.set_params(**params)

    def predict(self, X, return_std=False):
        """Predict.

        If `return_std` is set to False, only tree model prediction is returned.
        Return mean and predicted standard deviation, which is approximated 
        based on standard estimator specified, when `return_std` is set to True. 

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            where `n_samples` is the number of samples
            and `n_features` is the number of features.

        Returns
        -------
        mean or (mean, std): np.array, shape (n_rows, n_dims) 
            or tuple(np.array, np.array), depending on value of `return_std`.
        """
        mean = self.regressor_.predict(X)

        if return_std:
            std = self.std_estimator.predict(X)
            return mean, std

        # return the mean
        return mean

    def get_gbm_model(self):
        """
        Returns `GbmModel` instance of the tree model.

        Parameters
        ----------
        -

        Returns
        -------
        gbm_model : `GbmModel`
            ENTMOOT native tree model format to formulate optimization model
        """

        from entmoot.learning.lgbm_processing import order_tree_model_dict
        from entmoot.learning.gbm_model import GbmModel

        original_tree_model_dict = self.regressor_._Booster.dump_model()
        ordered_tree_model_dict = \
            order_tree_model_dict(original_tree_model_dict)
        gbm_model = GbmModel(ordered_tree_model_dict)
        return gbm_model






from sklearn.preprocessing import StandardScaler
from abc import ABC, abstractmethod

import numpy as np


class DistanceMetric(ABC):
    """Computes distances and defines the optimization model for both 
    exploration and penalty.

    Parameters
    ----------
    -

    Attributes
    ----------
    -
    """

    @abstractmethod
    def get_distance(self, x_left, x_right):
        """Compute the distance between `x_left` and `x_right` per row of x_left.

        Parameters
        ----------
        x_left : np.array, shape (n_rows, n_dims)
            Each row of `n_rows` is a reference point. Each dimension of `n_dim` 
            is the numerical value of a continuous variable.

        x_right : np.array, shape (n_dims,)
            Each dimension is the numerical value of a continuous variable.

        Returns
        -------
        dist : np.array, shape(n_rows,)
            Distance between `x_left` and `x_right`. If multiple rows are given 
            for `x_left`, a 1-dimensional array, i.e. `n_rows` > 1, is returned.
        """

        pass

    def get_max_space_scaled_dist(self, ref_points, x_means, x_stddev, model):
        # computes maximum distance in search space
        n_features = len(model._c_x)
        lb = np.asarray(model._c_x_lb)
        ub = np.asarray(model._c_x_ub)

        lb_std = np.divide(lb - x_means, x_stddev)
        ub_std = np.divide(ub - x_means, x_stddev)

        max_dist = self.get_distance(
            lb_std,
            ub_std
        )
        return max_dist

class SquaredEuclidean(DistanceMetric):
    """Computes distances and defines the optimization model for both 
    exploration and penalty. The distance metric used is the squared euclidean
    distance.

    Parameters
    ----------
    -

    Attributes
    ----------
    -
    """

    @staticmethod
    def get_distance(x_left, x_right):
        """Compute the distance between `x_left` and `x_right` per row of x_left.
        Here the squared euclidean distance is used.

        Parameters
        ----------
        x_left : np.array, shape (n_rows, n_dims)
            Each row of `n_rows` is a reference point. Each dimension of `n_dim` 
            is the numerical value of a continuous variable.

        x_right : np.array, shape (n_dims,)
            Each dimension is the numerical value of a continuous variable.

        Returns
        -------
        dist : np.array, shape(n_rows,)
            Distance between `x_left` and `x_right`. If multiple rows are given 
            for `x_left`, a 1-dimensional array, i.e. `n_rows` > 1, is returned.
        """

        if x_left.ndim == 1:
            dist = np.sum((x_left - x_right)**2)
        else:
            dist = np.sum((x_left - x_right)**2, axis=1)
        return dist

    def add_exploration_to_gurobi_model(self,
        ref_points, x_means, x_stddev, distance_bound, model):
        """Adds exploration constraints to a gurobi optimization model.
        Incentivizes solutions far away from reference points.

        Parameters
        ----------
        ref_points : np.array, shape (n_rows, n_dims)
            Each row of `n_rows` is a reference point. Each dimension of `n_dim` 
            is the numerical value of a continuous variable.

        x_means : np.array, shape (n_dims,)
            Each dimension is the mean value of a continuous variable used to 
            scale the data set.

        x_stddev : np.array, shape (n_dims,)
            Each dimension is the std value of a continuous variable used to 
            scale the data set.

        distance_bound : float
            Defines the maximum value that the exploration term can take.

        Returns
        -------
        -
        """

        from gurobipy import GRB, quicksum

        n_ref_points = len(ref_points)

        # variable alpha captures distance measure
        alpha_bound = distance_bound
        model._alpha = \
            model.addVar(
                lb=0, 
                ub=alpha_bound, 
                name="alpha", 
                vtype='C'
            )

        def distance_ref_point_i(model, xi_ref, x_mean, x_stddev):
            # function returns constraints capturing the standardized
            # exploration distance
            c_x = model._c_x
            alpha = model._alpha
            n_features = len(xi_ref)

            diff_to_ref_point_i = quicksum(
                ( (xi_ref[j] - (c_x[j]-x_mean[j]) / x_stddev[j]) * \
                    (xi_ref[j] - (c_x[j]-x_mean[j]) / x_stddev[j]) )
                for j in range(n_features)
            )
            
            return alpha <= diff_to_ref_point_i

        # add exploration distances as quadratic constraints to the model
        for i in range(n_ref_points):
            model.addQConstr(
                distance_ref_point_i(model, ref_points[i], x_means, x_stddev),
                name=f"std_const_{i}"
            )
        model.update()

    def add_penalty_to_gurobi_model(self,
        ref_points, x_means, x_stddev, model):
        """Adds penalty constraints to a gurobi optimization model.
        Incentivizes solutions close to reference points.

        Parameters
        ----------
        ref_points : np.array, shape (n_rows, n_dims)
            Each row of n_rows is a reference point. Each dimension of n_dim is 
            the numerical value of a continuous variable.

        x_means : np.array, shape (n_dims,)
            Each dimension is the mean value of a continuous variable used to 
            scale the data set.

        x_stddev : np.array, shape (n_dims,)
            Each dimension is the std value of a continuous variable used to 
            scale the data set.

        distance_bound : float
            Defines the maximum value that the exploration term can take.

        Returns
        -------
        -
        """

        from gurobipy import GRB, quicksum

        n_ref_points = len(ref_points)

        # big m is required to formulate the constraints
        model._big_m = \
            self.get_max_space_scaled_dist(ref_points, x_means, x_stddev, model)
        
        # binary variables b_ref correspond to active cluster centers
        model._b_ref = \
            model.addVars(
                range(n_ref_points),
                name="b_ref", 
                vtype=GRB.INTEGER
            )

        # variable alpha captures distance measure
        model._alpha = \
            model.addVar(
                ub=GRB.INFINITY,
                lb=0.0,
                name="alpha", 
                vtype='C'
            )

        def distance_ref_point_k(model, xk_ref, k_ref, x_mean, x_stddev):
            # function returns constraints capturing the standardized
            # penalty distance
            c_x = model._c_x
            b_ref = model._b_ref
            alpha = model._alpha
            n_features = len(xk_ref)

            diff_to_ref_point_k = quicksum(
                ( (xk_ref[j]  - (c_x[j]-x_mean[j]) / x_stddev[j]) * \
                    (xk_ref[j]  - (c_x[j]-x_mean[j]) / x_stddev[j]) )
                for j in range(n_features)
            )
            big_m_term = model._big_m*(1-b_ref[k_ref])
            return diff_to_ref_point_k <= alpha + big_m_term

        # add penalty distances as quadratic constraints to the model
        for k in range(n_ref_points):
            model.addQConstr(
                distance_ref_point_k(
                    model, 
                    ref_points[k], 
                    k, 
                    x_means, 
                    x_stddev
                ),
                name=f"std_const_{k}"
            )

        def sum_ref_point_vars(n_ref_points, model):
            return quicksum(model._b_ref[k] \
                for k in range(n_ref_points))== 1

        # add additional sum constraints forcing only one ref_point to
        # be active
        model.addConstr(
            sum_ref_point_vars(n_ref_points, model),
            name="std_ref_sum"
        )


class Manhattan(DistanceMetric):
    """Computes distances and defines the optimization model for both 
    exploration and penalty. The distance metric used is the manhattan
    distance.

    Parameters
    ----------
    -

    Attributes
    ----------
    -
    """

    @staticmethod
    def get_distance(x_left, x_right):
        """Compute the distance between `x_left` and `x_right` per row of 
        `x_left`. Here the manhattan distance is used.

        Parameters
        ----------
        x_left : np.array, shape (n_rows, n_dims)
            Each row of `n_rows` is a reference point. Each dimension of n_dim is 
            the numerical value of a continuous variable.

        x_right : np.array, shape (n_dims,)
            Each dimension is the numerical value of a continuous variable.

        Returns
        -------
        dist : np.array, shape(n_rows,)
            Distance between `x_left` and `x_right`. If multiple rows are given 
            for `x_left`, a 1-dimensional array, i.e. `n_rows` > 1, is returned.
        """

        if x_left.ndim == 1:
            dist = np.sum( np.abs(x_left - x_right) )
        else:
            dist = np.sum( np.abs(x_left - x_right), axis=1)
        return dist

    def add_exploration_to_gurobi_model(self,
        ref_points, x_means, x_stddev, distance_bound, model):
        """Adds exploration constraints to a gurobi optimization model.
        Incentivizes solutions far away from reference points.

        Parameters
        ----------
        ref_points : np.array, shape (n_rows, n_dims)
            Each row of `n_rows` is a reference point. Each dimension of `n_dim` 
            is the numerical value of a continuous variable.

        x_means : np.array, shape (n_dims,)
            Each dimension is the mean value of a continuous variable used to 
            scale the data set.

        x_stddev : np.array, shape (n_dims,)
            Each dimension is the std value of a continuous variable used to 
            scale the data set.

        distance_bound : float
            Defines the maximum value that the exploration term can take.

        Returns
        -------
        -
        """

        from gurobipy import GRB, quicksum

        n_ref_points = len(ref_points)

        # two sets of variables are used to capture positive and negative 
        # parts of manhattan distance
        model._c_x_aux_pos = \
            model.addVars(range(n_ref_points), range(model._n_feat), 
                        name="c_x_aux_pos", vtype='C')

        model._c_x_aux_neg = \
            model.addVars(range(n_ref_points), range(model._n_feat), 
                        name="c_x_aux_neg", vtype='C')
        
        # variable alpha captures distance measure
        alpha_bound = distance_bound
        model._alpha = \
            model.addVar(lb=0, 
                        ub=alpha_bound, 
                        name="alpha", vtype='C')

        def distance_ref_point_i_for_feat_j(
            model, xi_ref, i_ref, feat_j, x_mean, x_stddev):
            # function returns constraints capturing the standardized
            # exploration distance
            c_x = model._c_x

            diff_to_ref_point_i = \
                ( xi_ref[feat_j] - (c_x[feat_j]-x_mean[feat_j]) / \
                    x_stddev[feat_j] ) 
            return diff_to_ref_point_i == model._c_x_aux_pos[i_ref, feat_j] - \
                model._c_x_aux_neg[i_ref, feat_j]

        for i_ref in range(n_ref_points):
            for feat_j in range(model._n_feat):
                # add constraints to capture distances in variables
                # _c_x_aux_pos and _c_x_aux_neg
                model.addConstr(
                    distance_ref_point_i_for_feat_j(model, 
                        ref_points[i_ref], i_ref, feat_j, x_means, x_stddev),
                    name=f"std_const_feat_{feat_j}_{i_ref}"
                )
                # add sos constraints that allow only one of the +/- vars, 
                # i.e. _c_x_aux_pos / _c_x_aux_neg to be active
                model.addSOS(GRB.SOS_TYPE1, 
                    [
                        model._c_x_aux_pos[i_ref, feat_j], 
                        model._c_x_aux_neg[i_ref, feat_j]
                    ]
                )

            # add exploration distances as linear constraints to the model
            model.addConstr(
                model._alpha <= quicksum(
                    (model._c_x_aux_pos[i_ref, j] + model._c_x_aux_neg[i_ref, j])
                    for j in range(model._n_feat)
                ),
                name=f"alpha_sum"
            )
        model.update()

    def add_penalty_to_gurobi_model(self,
        ref_points, x_means, x_stddev, model):
        """Adds penalty constraints to a gurobi optimization model.
        Incentivizes solutions close to reference points.

        Parameters
        ----------
        ref_points : np.array, shape (n_rows, n_dims)
            Each row of `n_rows` is a reference point. Each dimension of `n_dim` 
            is the numerical value of a continuous variable.

        x_means : np.array, shape (n_dims,)
            Each dimension is the mean value of a continuous variable used to 
            scale the data set.

        x_stddev : np.array, shape (n_dims,)
            Each dimension is the std value of a continuous variable used to 
            scale the data set.

        distance_bound : float
            Defines the maximum value that the exploration term can take.

        Returns
        -------
        -
        """

        from gurobipy import GRB, quicksum

        n_ref_points = len(ref_points)

        # big m is required to formulate the constraints
        model._big_m = \
            self.get_max_space_scaled_dist(ref_points, x_means, x_stddev, model)
        
        # two sets of variables are used to capture positive and negative 
        # parts of manhattan distance
        model._c_x_aux_pos = \
            model.addVars(range(n_ref_points), range(model._n_feat), 
                        name="c_x_aux_pos", vtype='C')

        model._c_x_aux_neg = \
            model.addVars(range(n_ref_points), range(model._n_feat), 
                        name="c_x_aux_neg", vtype='C')
        
        # binary variables b_ref correspond to active cluster centers
        model._b_ref = \
            model.addVars(n_ref_points,
                        name="b_ref", vtype=GRB.INTEGER)

        # variable alpha captures distance measure
        model._alpha = \
            model.addVar(ub=GRB.INFINITY,
                        lb=0.0,
                        name="alpha", vtype='C')

        def distance_ref_point_i_for_feat_j(
            model, xi_ref, i_ref, feat_j, x_mean, x_stddev):
            # function returns constraints capturing the standardized
            # exploration distance
            c_x = model._c_x

            diff_to_ref_point_i = \
                ( xi_ref[feat_j] - (c_x[feat_j]-x_mean[feat_j]) / \
                    x_stddev[feat_j] ) 
            return diff_to_ref_point_i == model._c_x_aux_pos[i_ref, feat_j] - \
                model._c_x_aux_neg[i_ref, feat_j]

        for i_ref in range(n_ref_points):
            for feat_j in range(model._n_feat):
                # add constraints to capture distances in variables
                # _c_x_aux_pos and _c_x_aux_neg
                model.addConstr(
                    distance_ref_point_i_for_feat_j(
                        model, ref_points[i_ref], i_ref, feat_j, x_means, x_stddev),
                    name=f"std_const_feat_{feat_j}_{i_ref}"
                )
                # add sos constraints that allow only one of the +/- vars, 
                # i.e. _c_x_aux_pos / _c_x_aux_neg to be active
                model.addSOS(GRB.SOS_TYPE1, 
                    [model._c_x_aux_pos[i_ref, feat_j], model._c_x_aux_neg[i_ref, feat_j]]
                )

        # add penalty distances as linear constraints to the model
            model.addConstr(
                quicksum(
                    (model._c_x_aux_pos[i_ref, j] + model._c_x_aux_neg[i_ref, j])
                    for j in range(model._n_feat)
                ) <= model._alpha + model._big_m*(1-model._b_ref[i_ref]),
                name=f"alpha_sum"
            )
        model.update()

        def sum_ref_point_vars(n_ref_points, model):
            return quicksum(model._b_ref[k] for k in range(n_ref_points))== 1

        # add additional sum constraints forcing only one ref_point to
        # be active
        model.addConstr(
            sum_ref_point_vars(n_ref_points, model),
            name="std_ref_sum"
        )


class DistanceBasedStd(ABC):
    """Define a distance-based standard estimator.

    A `DistanceBasedStd` object is used to quantify model uncertainty based 
    on distance to reference points, e.g. data points. The underlying assumption 
    is that base estimator predictions are good close to training data.

    Use this class as a template if you want to develop your own distance-based
    measure.

    Parameters
    ----------
    metric : string
        Metric used to compute distances, e.g. squared euclidean, manhattan

    Attributes
    ----------
    Xi : list
        Points at which objective has been evaluated.
    yi : scalar
        Values of objective at corresponding points in `Xi`.
    cont_dist_metric : DistanceMetric
        Object used to compute distances between continuous variables.
    x_means : list
        Mean of attribute `Xi`.
    x_scaler : list
        Scalers of attribute `Xi`.
    Xi_standard : list
        Standardized `Xi` array.
    ref_points : list
        Points to which the distance is computed to estimate model uncertainty.
        Is different for all child classes.
    """
    def __init__(self, metric='sq_euclidean'):
        # define the distance metric for continuous variables
        if metric == 'sq_euclidean':
            from entmoot.learning.distance_based_std import SquaredEuclidean
            self.cont_dist_metric = SquaredEuclidean()

        elif metric == 'manhattan':
            from entmoot.learning.distance_based_std import Manhattan
            self.cont_dist_metric = Manhattan()

    def set_params(self,**kwargs):
        """Sets parameters related to distance-based standard estimator.

        Parameters
        ----------
        kwargs : dict
            Additional arguments to be passed to the standard estimator

        Returns
        -------
        -
        """
        pass

    def update(self, Xi, yi):
        """Update available data points which is usually done after every
        iteration.

        Parameters
        ----------
        Xi : list
            Points at which objective has been evaluated.
        yi : scalar
            Values of objective at corresponding points in `Xi`.

        Returns
        -------
        -
        """
        # update data set attributes
        self.Xi = Xi
        self.yi = yi

        # compute mean and scaler of data set
        standard_scaler = StandardScaler()
        projected_features = standard_scaler.fit_transform(self.Xi)
        self.x_means = standard_scaler.mean_
        self.x_scalers = standard_scaler.scale_

        # standardize dataset
        self.Xi_standard = self.standardize_with_Xi(self.Xi)

    def standardize_with_Xi(self, X):
        """Standardize given input `X` based on attribute `Xi`.

        Parameters
        ----------
        X : numpy array, shape (n_rows, n_dims)
            Each row of n_rows is a point in `X`.
            Points which are standardized based on `Xi`

        Returns
        -------
        x_standard : numpy array, shape (n_rows, n_dims)
            Standardized array of `X`
        """
        x_standard = np.divide(X - self.x_means, self.x_scalers)
        return x_standard

    def get_closest_point_distance(self, X):
        """Get distance to point of attribute `ref_points` which is closest to 
        point given as parameter `X`.

        Parameters
        ----------
        X : numpy array, shape (n_dims,)
            Point to which the distance of closest reference point is 
            computed

        Returns
        -------
        dist : numpy array, shape (1, )
            Returns distance to closest `ref_point`.
        """
        ref_points = np.asarray(self.ref_points)

        x_standard = self.standardize_with_Xi(X)
        x_standard = np.asarray(x_standard)

        dist_cont = \
            self.cont_dist_metric.get_distance(ref_points,x_standard)
        
        return np.min(dist_cont)

    def predict(self, X):
        """Predict standard estimate at location `X`.

        Parameters
        ----------
        X : numpy array, shape (n_rows, n_dims)
            Points at which the standard estimator is evaluated.

        Returns
        -------
        dist : numpy array, shape (n_rows,)
            Returns distances to closest `ref_point` for every point per row
            in `n_rows`.
        """
        dist = np.empty([X.shape[0],])

        for row_res,Xi in enumerate(X):
            ref_distance = self.get_closest_point_distance(Xi)
            dist[row_res] = ref_distance     
        return dist

    @abstractmethod
    def add_to_gurobi_model(self,model):
        """Add standard estimator to gurobi model. Model details are 
        implemented in child class.

        Parameters
        ----------
        model : gurobipy.Model,
            Model to which the standard estimator is added.
        """
        pass

    @abstractmethod
    def get_gurobi_obj(self,model):
        """Get contribution of standard estimator to gurobi model objective
        function.

        Parameters
        ----------
        model : gurobipy.Model,
            Model to which the standard estimator is added.

        s
        ----------
        alpha : gurobipy.Var,
            Model variable that takes the value of the uncertainty measure.
        """
        pass


class DistanceBasedExploration(DistanceBasedStd):
    """Defines a child class based on `DistanceBasedStd`. Exploration 
    refers to how the distance measure contributes to the acquisition
    function. Exploration refers to incentivizing distance to reference points
    leading to a negative contribution of the distance measure to the objective
    function.
    
    Parameters
    ----------
    metric : string
        Metric used to compute distances, e.g. squared euclidean, manhattan
    zeta : scalar
        Coefficient determining how the distance measure is bounded

    Attributes
    ----------
    Xi : list
        Points at which objective has been evaluated.
    yi : scalar
        Values of objective at corresponding points in `Xi`.
    cont_dist_metric : DistanceMetric
        Object used to compute distances between continuous variables.
    x_means : list
        Mean of attribute `Xi`.
    x_scaler : list
        Scalers of attribute `Xi`.
    Xi_standard : list
        Standardized `Xi` array.
    ref_points : list
        `ref_points` standardized to which the distance measure is computed.
    ref_points_unscaled : list
        Unscaled `ref_points` to which the distance measure is computed.
    distance_bound : scalar
        Bound of exploration measure to prohibit over-exploration
    """
    def __init__(self,
        metric="sq_euclidean",
        zeta=0.5):

        super().__init__(metric)
        self.zeta = zeta

    def set_params(self,**kwargs):
        """Sets parameters related to distance-based standard estimator.

        Parameters
        ----------
        kwargs : dict
            Additional arguments to be passed to the standard estimator

        Returns
        -------
        -
        """
        zeta = kwargs.get("zeta", 0.5)
        self.zeta = zeta

    def update(self, Xi, yi):
        """Update available data points which is usually done after every
        iteration.

        Parameters
        ----------
        Xi : list
            Points at which objective has been evaluated.
        yi : scalar
            Values of objective at corresponding points in `Xi`.

        Returns
        -------
        -
        """
        super().update(Xi, yi)

        self.ref_points_unscaled = self.Xi

        self.ref_points = self.Xi_standard

        # compute upper bound of uncertainty
        y_var = np.var(yi)
        self.distance_bound = abs(self.zeta*y_var)

    def predict(self, X):
        """Predict standard estimate at location `X`. By default `dist` is
        bounded by attribute `distance_bound`.

        Parameters
        ----------
        X : numpy array, shape (n_rows, n_dims)
            Points at which the standard estimator is evaluated.

        Returns
        -------
        dist : numpy array, shape (n_rows,)
            Returns distances to closest `ref_point` for every point per row
            in `n_rows`.
        """
        dist = super().predict(X)

        # prediction has max out at `distance_bound`
        dist[dist > self.distance_bound] = self.distance_bound
        return dist

    def get_gurobi_obj(self, model):
        """Get contribution of standard estimator to gurobi model objective
        function.

        Parameters
        ----------
        model : gurobipy.Model,
            Model to which the standard estimator is added.

        Returns
        -------
        alpha : gurobipy.Var,
            Model variable that takes the value of the uncertainty measure.
        """
        # negative contributation of alpha requires non-convex flag in gurobi.
        model.Params.NonConvex=2
        return -model._alpha

    def add_to_gurobi_model(self,model):
        """Add standard estimator to gurobi model. Model details are 
        implemented in child class.

        Parameters
        ----------
        model : gurobipy.Model,
            Model to which the standard estimator is added.
        """
        self.cont_dist_metric.add_exploration_to_gurobi_model(
            self.ref_points, 
            self.x_means, 
            self.x_scalers, 
            self.distance_bound, 
            model
        )

class DistanceBasedPenalty(DistanceBasedStd):
    """Defines a child class based on `DistanceBasedStd`. Penalty 
    refers to how the distance measure contributes to the acquisition
    function. Penalty refers to penalizing distance to reference points
    leading to a positive contribution of the distance measure to the objective
    function.
    
    Parameters
    ----------
    metric : string
        Metric used to compute distances, e.g. squared euclidean, manhattan

    Attributes
    ----------
    Xi : list
        Points at which objective has been evaluated.
    yi : scalar
        Values of objective at corresponding points in `Xi`.
    cont_dist_metric : DistanceMetric
        Object used to compute distances between continuous variables.
    x_means : list
        Mean of attribute `Xi`.
    x_scaler : list
        Scalers of attribute `Xi`.
    Xi_standard : list
        Standardized `Xi` array.
    ref_points : list
        `ref_points` reference to which the distance measure is computed
    n_ref_points : scalar
        length of `ref_points"""

    def __init__(self,
        metric="sq_euclidean"):

        super().__init__(metric)

    def set_params(self,**kwargs):
        """Sets parameters related to distance-based standard estimator.

        Parameters
        ----------
        kwargs : dict
            Additional arguments to be passed to the standard estimator

        Returns
        -------
        -
        """
        pass

    def update(self, Xi, yi):
        """Update available data points which is usually done after every
        iteration.

        Parameters
        ----------
        Xi : list
            Points at which objective has been evaluated.
        yi : scalar
            Values of objective at corresponding points in `Xi`.

        Returns
        -------
        -
        """
        super().update(Xi, yi)

        self.ref_points_unscaled = self.Xi

        self.ref_points = self.Xi_standard  


    def predict(self, X):
        """Predict standard estimate at location `X`. Sign of `dist` is negative
        because it contributes as a penalty.

        Parameters
        ----------
        X : numpy array, shape (n_rows, n_dims)
            Points at which the standard estimator is evaluated.

        Returns
        -------
        dist : numpy array, shape (n_rows,)
            Returns distances to closest `ref_point` for every point per row
            in `n_rows`.
        """
        dist = super().predict(X)
        return -dist

    def get_gurobi_obj(self, model):
        """Get contribution of standard estimator to gurobi model objective
        function.

        Parameters
        ----------
        model : gurobipy.Model,
            Model to which the standard estimator is added.

        Returns
        -------
        alpha : gurobipy.Var,
            Model variable that takes the value of the uncertainty measure.
        """
        return model._alpha

    def add_to_gurobi_model(self,model):
        """Add standard estimator to gurobi model. Model details are 
        implemented in child class.

        Parameters
        ----------
        model : gurobipy.Model,
            Model to which the standard estimator is added.
        """
        self.cont_dist_metric.add_penalty_to_gurobi_model(
            self.ref_points, 
            self.x_means, 
            self.x_scalers, 
            model
        )


"""
Copyright (c) 2016-2020 The scikit-optimize developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

NOTE: Changes were made to the scikit-optimize source code included here. 
For the most recent version of scikit-optimize we refer to:
https://github.com/scikit-optimize/scikit-optimize/

Copyright (c) 2019-2020 Alexander Thebelt.
"""

import numpy as np
import warnings
import sys

def _gaussian_acquisition(X, model, y_opt=None, acq_func="LCB",
                          acq_func_kwargs=None):
    """
    Wrapper so that the output of this function can be
    directly passed to a minimizer.
    """
    # Check inputs
    X = np.asarray(X)
    if X.ndim != 2:
        raise ValueError("X is {}-dimensional, however,"
                         " it must be 2-dimensional.".format(X.ndim))

    if acq_func_kwargs is None:
        acq_func_kwargs = dict()
    xi = acq_func_kwargs.get("xi", 0.01)
    kappa = acq_func_kwargs.get("kappa", 1.96)

    # Evaluate acquisition function
    if acq_func == "LCB":
        acq_vals = gaussian_lcb(X, model, kappa)
    else:
        raise ValueError("Acquisition function not implemented.")

    return acq_vals

def gaussian_lcb(X, model, kappa=1.96, return_grad=False):
    """
    Use the lower confidence bound to estimate the acquisition
    values.

    The trade-off between exploitation and exploration is left to
    be controlled by the user through the parameter ``kappa``.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Values where the acquisition function should be computed.

    model : sklearn estimator that implements predict with ``return_std``
        The fit estimator that approximates the function through the
        method ``predict``.
        It should have a ``return_std`` parameter that returns the standard
        deviation.

    kappa : float, default 1.96 or 'inf'
        Controls how much of the variance in the predicted values should be
        taken into account. If set to be very high, then we are favouring
        exploration over exploitation and vice versa.
        If set to 'inf', the acquisition function will only use the variance
        which is useful in a pure exploration setting.
        Useless if ``method`` is not set to "LCB".

    Returns
    -------
    values : array-like, shape (X.shape[0],)
        Acquisition function values computed at X.

    """
    # Compute posterior
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        mu, std = model.predict(X, return_std=True)

        if kappa == "inf":
            return -std

        return mu - kappa * std




"""
Copyright (c) 2016-2020 The scikit-optimize developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import numbers
import numpy as np
import yaml

from scipy.stats.distributions import randint
from scipy.stats.distributions import rv_discrete
from scipy.stats.distributions import uniform

from sklearn.utils import check_random_state
from sklearn.utils.fixes import sp_version

from entmoot.space.transformers import CategoricalEncoder
from entmoot.space.transformers import StringEncoder
from entmoot.space.transformers import LabelEncoder
from entmoot.space.transformers import Normalize
from entmoot.space.transformers import Identity
from entmoot.space.transformers import LogN
from entmoot.space.transformers import Pipeline


# helper class to be able to print [1, ..., 4] instead of [1, '...', 4]
class _Ellipsis:
    def __repr__(self):
        return '...'


def _transpose_list_array(x):
    """Transposes a list matrix
    """

    n_dims = len(x)
    assert n_dims > 0
    n_samples = len(x[0])
    rows = [None] * n_samples
    for i in range(n_samples):
        r = [None] * n_dims
        for j in range(n_dims):
            r[j] = x[j][i]
        rows[i] = r
    return rows


def check_dimension(dimension, transform=None):
    """Turn a provided dimension description into a dimension object.

    Checks that the provided dimension falls into one of the
    supported types. For a list of supported types, look at
    the documentation of ``dimension`` below.

    If ``dimension`` is already a ``Dimension`` instance, return it.

    Parameters
    ----------
    dimension : Dimension
        Search space Dimension.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

    transform : "identity", "normalize", "string", "label", "onehot" optional
        - For `Categorical` dimensions, the following transformations are
          supported.

          - "onehot" (default) one-hot transformation of the original space.
          - "label" integer transformation of the original space
          - "string" string transformation of the original space.
          - "identity" same as the original space.

        - For `Real` and `Integer` dimensions, the following transformations
          are supported.

          - "identity", (default) the transformed space is the same as the
            original space.
          - "normalize", the transformed space is scaled to be between 0 and 1.

    Returns
    -------
    dimension : Dimension
        Dimension instance.
    """
    if isinstance(dimension, Dimension):
        return dimension

    if not isinstance(dimension, (list, tuple, np.ndarray)):
        raise ValueError("Dimension has to be a list or tuple.")

    # A `Dimension` described by a single value is assumed to be
    # a `Categorical` dimension. This can be used in `BayesSearchCV`
    # to define subspaces that fix one value, e.g. to choose the
    # model type, see "sklearn-gridsearchcv-replacement.py"
    # for examples.
    if len(dimension) == 1:
        return Categorical(dimension, transform=transform)

    if len(dimension) == 2:
        if any([isinstance(d, (str, bool)) or isinstance(d, np.bool_)
                for d in dimension]):
            return Categorical(dimension, transform=transform)
        elif all([isinstance(dim, numbers.Integral) for dim in dimension]):
            return Integer(*dimension, transform=transform)
        elif any([isinstance(dim, numbers.Real) for dim in dimension]):
            return Real(*dimension, transform=transform)
        else:
            raise ValueError("Invalid dimension {}. Read the documentation for"
                             " supported types.".format(dimension))

    if len(dimension) == 3:
        if (any([isinstance(dim, int) for dim in dimension[:2]]) and
                dimension[2] in ["uniform", "log-uniform"]):
            return Integer(*dimension, transform=transform)
        elif (any([isinstance(dim, (float, int)) for dim in dimension[:2]]) and
              dimension[2] in ["uniform", "log-uniform"]):
            return Real(*dimension, transform=transform)
        else:
            return Categorical(dimension, transform=transform)

    if len(dimension) == 4:
        if (any([isinstance(dim, int) for dim in dimension[:2]]) and
                dimension[2] == "log-uniform" and isinstance(dimension[3],
                                                             int)):
            return Integer(*dimension, transform=transform)
        elif (any([isinstance(dim, (float, int)) for dim in dimension[:2]]) and
              dimension[2] == "log-uniform" and isinstance(dimension[3], int)):
            return Real(*dimension, transform=transform)

    if len(dimension) > 3:
        return Categorical(dimension, transform=transform)

    raise ValueError("Invalid dimension {}. Read the documentation for "
                     "supported types.".format(dimension))


class Dimension(object):
    """Base class for search space dimensions."""

    prior = None

    def rvs(self, n_samples=1, random_state=None):
        """Draw random samples.

        Parameters
        ----------
        n_samples : int or None
            The number of samples to be drawn.

        random_state : int, RandomState instance, or None (default)
            Set random state to something other than None for reproducible
            results.
        """
        rng = check_random_state(random_state)
        samples = self._rvs.rvs(size=n_samples, random_state=rng)
        return self.inverse_transform(samples)

    def transform(self, X):
        """Transform samples form the original space to a warped space."""
        return self.transformer.transform(X)

    def inverse_transform(self, Xt):
        """Inverse transform samples from the warped space back into the
           original space.
        """
        return self.transformer.inverse_transform(Xt)

    def set_transformer(self):
        raise NotImplementedError

    @property
    def size(self):
        return 1

    @property
    def transformed_size(self):
        return 1

    @property
    def bounds(self):
        raise NotImplementedError

    @property
    def is_constant(self):
        raise NotImplementedError

    @property
    def transformed_bounds(self):
        raise NotImplementedError

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if isinstance(value, str) or value is None:
            self._name = value
        else:
            raise ValueError("Dimension's name must be either string or None.")


def _uniform_inclusive(loc=0.0, scale=1.0):
    # like scipy.stats.distributions but inclusive of `high`
    # XXX scale + 1. might not actually be a float after scale if
    # XXX scale is very large.
    return uniform(loc=loc, scale=np.nextafter(scale, scale + 1.))


class Real(Dimension):
    """Search space dimension that can take on any real value.

    Parameters
    ----------
    low : float
        Lower bound (inclusive).

    high : float
        Upper bound (inclusive).

    prior : "uniform" or "log-uniform", default="uniform"
        Distribution to use when sampling random points for this dimension.

        - If `"uniform"`, points are sampled uniformly between the lower
          and upper bounds.
        - If `"log-uniform"`, points are sampled uniformly between
          `log(lower, base)` and `log(upper, base)` where log
          has base `base`.

    base : int
        The logarithmic base to use for a log-uniform prior.
        - Default 10, otherwise commonly 2.

    transform : "identity", "normalize", optional
        The following transformations are supported.

        - "identity", (default) the transformed space is the same as the
          original space.
        - "normalize", the transformed space is scaled to be between
          0 and 1.

    name : str or None
        Name associated with the dimension, e.g., "learning rate".

    dtype : str or dtype, default=np.float
        float type which will be used in inverse_transform,
        can be float.

    """
    def __init__(self, low, high, prior="uniform", base=10, transform=None,
                 name=None, dtype=np.float):
        if high <= low:
            raise ValueError("the lower bound {} has to be less than the"
                             " upper bound {}".format(low, high))
        self.low = low
        self.high = high
        self.prior = prior
        self.base = base
        self.log_base = np.log10(base)
        self.name = name
        self.dtype = dtype
        self._rvs = None
        self.transformer = None
        self.transform_ = transform
        if isinstance(self.dtype, str) and self.dtype\
                not in ['float', 'float16', 'float32', 'float64']:
            raise ValueError("dtype must be 'float', 'float16', 'float32'"
                             "or 'float64'"
                             " got {}".format(self.dtype))
        elif isinstance(self.dtype, type) and self.dtype\
                not in [float, np.float, np.float16, np.float32, np.float64]:
            raise ValueError("dtype must be float, np.float"
                             " got {}".format(self.dtype))

        if transform is None:
            transform = "identity"
        self.set_transformer(transform)

    def set_transformer(self, transform="identitiy"):
        """Define rvs and transformer spaces.

        Parameters
        ----------
        transform : str
           Can be 'normalize' or 'identity'

        """
        self.transform_ = transform

        if self.transform_ not in ["normalize", "identity"]:
            raise ValueError("transform should be 'normalize' or 'identity'"
                             " got {}".format(self.transform_))

        # XXX: The _rvs is for sampling in the transformed space.
        # The rvs on Dimension calls inverse_transform on the points sampled
        # using _rvs
        if self.transform_ == "normalize":
            # set upper bound to next float after 1. to make the numbers
            # inclusive of upper edge
            self._rvs = _uniform_inclusive(0., 1.)
            if self.prior == "uniform":
                self.transformer = Pipeline(
                    [Identity(), Normalize(self.low, self.high)])
            else:
                self.transformer = Pipeline(
                    [LogN(self.base),
                     Normalize(np.log10(self.low) / self.log_base,
                               np.log10(self.high) / self.log_base)]
                )
        else:
            if self.prior == "uniform":
                self._rvs = _uniform_inclusive(self.low, self.high - self.low)
                self.transformer = Identity()
            else:
                self._rvs = _uniform_inclusive(
                    np.log10(self.low) / self.log_base,
                    np.log10(self.high) / self.log_base -
                    np.log10(self.low) / self.log_base)
                self.transformer = LogN(self.base)

    def __eq__(self, other):
        return (type(self) is type(other) and
                np.allclose([self.low], [other.low]) and
                np.allclose([self.high], [other.high]) and
                self.prior == other.prior and
                self.transform_ == other.transform_)

    def __repr__(self):
        return "Real(low={}, high={}, prior='{}', transform='{}')".format(
            self.low, self.high, self.prior, self.transform_)

    def inverse_transform(self, Xt):
        """Inverse transform samples from the warped space back into the
           original space.
        """

        inv_transform = super(Real, self).inverse_transform(Xt)
        if isinstance(inv_transform, list):
            inv_transform = np.array(inv_transform)
        inv_transform = np.clip(inv_transform,
                                self.low, self.high).astype(self.dtype)
        if self.dtype == float or self.dtype == 'float':
            # necessary, otherwise the type is converted to a numpy type
            return getattr(inv_transform, "tolist", lambda: value)()
        else:
            return inv_transform

    @property
    def bounds(self):
        return (self.low, self.high)

    @property
    def is_constant(self):
        return self.low == self.high

    def __contains__(self, point):
        if isinstance(point, list):
            point = np.array(point)
        return self.low <= point <= self.high

    @property
    def transformed_bounds(self):
        if self.transform_ == "normalize":
            return 0.0, 1.0
        else:
            if self.prior == "uniform":
                return self.low, self.high
            else:
                return np.log10(self.low), np.log10(self.high)

    def distance(self, a, b):
        """Compute distance between point `a` and `b`.

        Parameters
        ----------
        a : float
            First point.

        b : float
            Second point.
        """
        if not (a in self and b in self):
            raise RuntimeError("Can only compute distance for values within "
                               "the space, not %s and %s." % (a, b))
        return abs(a - b)


class Integer(Dimension):
    """Search space dimension that can take on integer values.

    Parameters
    ----------
    low : int
        Lower bound (inclusive).

    high : int
        Upper bound (inclusive).

    prior : "uniform" or "log-uniform", default="uniform"
        Distribution to use when sampling random integers for
        this dimension.

        - If `"uniform"`, integers are sampled uniformly between the lower
          and upper bounds.
        - If `"log-uniform"`, integers are sampled uniformly between
          `log(lower, base)` and `log(upper, base)` where log
          has base `base`.

    base : int
        The logarithmic base to use for a log-uniform prior.

        - Default 10, otherwise commonly 2.

    transform : "identity", "normalize", optional
        The following transformations are supported.

        - "identity", (default) the transformed space is the same as the
          original space.
        - "normalize", the transformed space is scaled to be between
          0 and 1.

    name : str or None
        Name associated with dimension, e.g., "number of trees".

    dtype : str or dtype, default=np.int64
        integer type which will be used in inverse_transform,
        can be int, np.int16, np.uint32, np.int32, np.int64 (default).
        When set to int, `inverse_transform` returns a list instead of
        a numpy array

    """
    def __init__(self, low, high, prior="uniform", base=10, transform=None,
                 name=None, dtype=np.int64):
        if high <= low:
            raise ValueError("the lower bound {} has to be less than the"
                             " upper bound {}".format(low, high))
        self.low = low
        self.high = high
        self.prior = prior
        self.base = base
        self.log_base = np.log10(base)
        self.name = name
        self.dtype = dtype
        self.transform_ = transform
        self._rvs = None
        self.transformer = None

        if isinstance(self.dtype, str) and self.dtype\
            not in ['int', 'int8', 'int16', 'int32', 'int64',
                    'uint8', 'uint16', 'uint32', 'uint64']:
            raise ValueError("dtype must be 'int', 'int8', 'int16',"
                             "'int32', 'int64', 'uint8',"
                             "'uint16', 'uint32', or"
                             "'uint64', but got {}".format(self.dtype))
        elif isinstance(self.dtype, type) and self.dtype\
                not in [int, np.int8, np.int16, np.int32, np.int64,
                        np.uint8, np.uint16, np.uint32, np.uint64]:
            raise ValueError("dtype must be 'int', 'np.int8', 'np.int16',"
                             "'np.int32', 'np.int64', 'np.uint8',"
                             "'np.uint16', 'np.uint32', or"
                             "'np.uint64', but got {}".format(self.dtype))

        if transform is None:
            transform = "identity"
        self.set_transformer(transform)

    def set_transformer(self, transform="identitiy"):
        """Define _rvs and transformer spaces.

        Parameters
        ----------
        transform : str
           Can be 'normalize' or 'identity'

        """
        self.transform_ = transform

        if transform not in ["normalize", "identity"]:
            raise ValueError("transform should be 'normalize' or 'identity'"
                             " got {}".format(self.transform_))

        if self.transform_ == "normalize":
            self._rvs = _uniform_inclusive(0.0, 1.0)
            if self.prior == "uniform":
                self.transformer = Pipeline(
                    [Identity(), Normalize(self.low, self.high, is_int=True)])
            else:

                self.transformer = Pipeline(
                    [LogN(self.base),
                     Normalize(np.log10(self.low) / self.log_base,
                               np.log10(self.high) / self.log_base)]
                )
        else:
            if self.prior == "uniform":
                self._rvs = randint(self.low, self.high + 1)
                self.transformer = Identity()
            else:
                self._rvs = _uniform_inclusive(
                    np.log10(self.low) / self.log_base,
                    np.log10(self.high) / self.log_base -
                    np.log10(self.low) / self.log_base)
                self.transformer = LogN(self.base)

    def __eq__(self, other):
        return (type(self) is type(other) and
                np.allclose([self.low], [other.low]) and
                np.allclose([self.high], [other.high]))

    def __repr__(self):
        return "Integer(low={}, high={}, prior='{}', transform='{}')".format(
            self.low, self.high, self.prior, self.transform_)

    def inverse_transform(self, Xt):
        """Inverse transform samples from the warped space back into the
           original space.
        """
        # The concatenation of all transformed dimensions makes Xt to be
        # of type float, hence the required cast back to int.
        inv_transform = super(Integer, self).inverse_transform(Xt)
        if isinstance(inv_transform, list):
            inv_transform = np.array(inv_transform)
        inv_transform = np.clip(inv_transform,
                                self.low, self.high)
        if self.dtype == int or self.dtype == 'int':
            # necessary, otherwise the type is converted to a numpy type
            return getattr(np.round(inv_transform).astype(self.dtype),
                           "tolist", lambda: value)()
        else:
            return np.round(inv_transform).astype(self.dtype)

    @property
    def bounds(self):
        return (self.low, self.high)

    @property
    def is_constant(self):
        return self.low == self.high

    def __contains__(self, point):
        if isinstance(point, list):
            point = np.array(point)
        return self.low <= point <= self.high

    @property
    def transformed_bounds(self):
        if self.transform_ == "normalize":
            return 0., 1.
        else:
            return (self.low, self.high)

    def distance(self, a, b):
        """Compute distance between point `a` and `b`.

        Parameters
        ----------
        a : int
            First point.

        b : int
            Second point.
        """
        if not (a in self and b in self):
            raise RuntimeError("Can only compute distance for values within "
                               "the space, not %s and %s." % (a, b))
        return abs(a - b)


class Categorical(Dimension):
    """Search space dimension that can take on categorical values.

    Parameters
    ----------
    categories : list, shape=(n_categories,)
        Sequence of possible categories.

    prior : list, shape=(categories,), default=None
        Prior probabilities for each category. By default all categories
        are equally likely.

    transform : "onehot", "string", "identity", "label", default="onehot"

        - "identity", the transformed space is the same as the original
          space.
        - "string",  the transformed space is a string encoded
          representation of the original space.
        - "label", the transformed space is a label encoded
          representation (integer) of the original space.
        - "onehot", the transformed space is a one-hot encoded
          representation of the original space.

    name : str or None
        Name associated with dimension, e.g., "colors".

    """
    def __init__(self, categories, prior=None, transform=None, name=None):
        self.categories = tuple(categories)

        self.name = name

        if transform is None:
            transform = "onehot"
        self.transform_ = transform
        self.transformer = None
        self._rvs = None
        self.prior = prior

        if prior is None:
            self.prior_ = np.tile(1. / len(self.categories),
                                  len(self.categories))
        else:
            self.prior_ = prior
        self.set_transformer(transform)

    def set_transformer(self, transform="onehot"):
        """Define _rvs and transformer spaces.

        Parameters
        ----------
        transform : str
           Can be 'normalize', 'onehot', 'string', 'label', or 'identity'

        """
        self.transform_ = transform
        if transform not in ["identity", "onehot", "string", "normalize",
                             "label"]:
            raise ValueError("Expected transform to be 'identity', 'string',"
                             "'label' or 'onehot' got {}".format(transform))
        if transform == "onehot":
            self.transformer = CategoricalEncoder()
            self.transformer.fit(self.categories)
        elif transform == "string":
            self.transformer = StringEncoder()
            self.transformer.fit(self.categories)
        elif transform == "label":
            self.transformer = LabelEncoder()
            self.transformer.fit(self.categories)
        elif transform == "normalize":
            self.transformer = Pipeline(
                [LabelEncoder(list(self.categories)),
                 Normalize(0, len(self.categories) - 1, is_int=True)])
        else:
            self.transformer = Identity()
            self.transformer.fit(self.categories)
        if transform == "normalize":
            self._rvs = _uniform_inclusive(0.0, 1.0)
        else:
            # XXX check that sum(prior) == 1
            self._rvs = rv_discrete(
                values=(range(len(self.categories)), self.prior_)
            )

    def __eq__(self, other):
        return (type(self) is type(other) and
                self.categories == other.categories and
                np.allclose(self.prior_, other.prior_))

    def __repr__(self):
        if len(self.categories) > 7:
            cats = self.categories[:3] + (_Ellipsis(),) + self.categories[-3:]
        else:
            cats = self.categories

        if self.prior is not None and len(self.prior) > 7:
            prior = self.prior[:3] + [_Ellipsis()] + self.prior[-3:]
        else:
            prior = self.prior

        return "Categorical(categories={}, prior={})".format(cats, prior)

    def inverse_transform(self, Xt):
        """Inverse transform samples from the warped space back into the
           original space.
        """
        # The concatenation of all transformed dimensions makes Xt to be
        # of type float, hence the required cast back to int.
        inv_transform = super(Categorical, self).inverse_transform(Xt)
        if isinstance(inv_transform, list):
            inv_transform = np.array(inv_transform)
        return inv_transform

    def rvs(self, n_samples=None, random_state=None):
        choices = self._rvs.rvs(size=n_samples, random_state=random_state)

        if isinstance(choices, numbers.Integral):
            return self.categories[choices]
        elif self.transform_ == "normalize" and isinstance(choices, float):
            return self.inverse_transform([(choices)])
        elif self.transform_ == "normalize":
            return self.inverse_transform(list(choices))
        else:
            return [self.categories[c] for c in choices]

    @property
    def transformed_size(self):
        if self.transform_ == "onehot":
            size = len(self.categories)
            # when len(categories) == 2, CategoricalEncoder outputs a
            # single value
            return size if size != 2 else 1
        return 1

    @property
    def bounds(self):
        return self.categories

    @property
    def is_constant(self):
        return len(self.categories) <= 1

    def __contains__(self, point):
        return point in self.categories

    @property
    def transformed_bounds(self):
        if self.transformed_size == 1:
            return 0.0, 1.0
        else:
            return [(0.0, 1.0) for i in range(self.transformed_size)]

    def distance(self, a, b):
        """Compute distance between category `a` and `b`.

        As categories have no order the distance between two points is one
        if a != b and zero otherwise.

        Parameters
        ----------
        a : category
            First category.

        b : category
            Second category.
        """
        if not (a in self and b in self):
            raise RuntimeError("Can only compute distance for values within"
                               " the space, not {} and {}.".format(a, b))
        return 1 if a != b else 0


class Space(object):
    """Initialize a search space from given specifications.

    Parameters
    ----------
    dimensions : list, shape=(n_dims,)
        List of search space dimensions.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

        .. note::
            The upper and lower bounds are inclusive for `Integer`
            dimensions.
    """
    def __init__(self, dimensions):
        self.dimensions = [check_dimension(dim) for dim in dimensions]

    def __eq__(self, other):
        return all([a == b for a, b in zip(self.dimensions, other.dimensions)])

    def __repr__(self):
        if len(self.dimensions) > 31:
            dims = self.dimensions[:15] + [_Ellipsis()] + self.dimensions[-15:]
        else:
            dims = self.dimensions
        return "Space([{}])".format(',\n       '.join(map(str, dims)))

    def __iter__(self):
        return iter(self.dimensions)

    @property
    def dimension_names(self):
        """
        Names of all the dimensions in the search-space.
        """
        index = 0
        names = []
        for dim in self.dimensions:
            if dim.name is None:
                names.append("X_%d" % index)
            else:
                names.append(dim.name)
            index += 1
        return names

    @property
    def is_real(self):
        """
        Returns true if all dimensions are Real
        """
        return all([isinstance(dim, Real) for dim in self.dimensions])

    @classmethod
    def from_yaml(cls, yml_path, namespace=None):
        """Create Space from yaml configuration file

        Parameters
        ----------
        yml_path : str
            Full path to yaml configuration file, example YaML below:
            Space:

            - Integer:
              low: -5
              high: 5
            - Categorical:
              categories:
              - a
              - b
            - Real:
              low: 1.0
              high: 5.0
              prior: log-uniform

        namespace : str, default=None
           Namespace within configuration file to use, will use first
           namespace if not provided

        Returns
        -------
        space : Space
           Instantiated Space object

        """
        with open(yml_path, 'rb') as f:
            config = yaml.safe_load(f)

        dimension_classes = {'real': Real,
                             'integer': Integer,
                             'categorical': Categorical}

        # Extract space options for configuration file
        if isinstance(config, dict):
            if namespace is None:
                options = next(iter(config.values()))
            else:
                options = config[namespace]
        elif isinstance(config, list):
            options = config
        else:
            raise TypeError('YaML does not specify a list or dictionary')

        # Populate list with Dimension objects
        dimensions = []
        for option in options:
            key = next(iter(option.keys()))
            # Make configuration case insensitive
            dimension_class = key.lower()
            values = {k.lower(): v for k, v in option[key].items()}
            if dimension_class in dimension_classes:
                # Instantiate Dimension subclass and add it to the list
                dimension = dimension_classes[dimension_class](**values)
                dimensions.append(dimension)

        space = cls(dimensions=dimensions)

        return space

    def rvs(self, n_samples=1, random_state=None):
        """Draw random samples.

        The samples are in the original space. They need to be transformed
        before being passed to a model or minimizer by `space.transform()`.

        Parameters
        ----------
        n_samples : int, default=1
            Number of samples to be drawn from the space.

        random_state : int, RandomState instance, or None (default)
            Set random state to something other than None for reproducible
            results.

        Returns
        -------
        points : list of lists, shape=(n_points, n_dims)
           Points sampled from the space.
        """
        rng = check_random_state(random_state)

        # Draw
        columns = []

        for dim in self.dimensions:
            columns.append(dim.rvs(n_samples=n_samples, random_state=rng))
        
        # Transpose
        return _transpose_list_array(columns)

    def set_transformer(self, transform):
        """Sets the transformer of all dimension objects to `transform`

        Parameters
        ----------
        transform : str or list of str
           Sets all transformer,, when `transform`  is a string.
           Otherwise, transform must be a list with strings with
           the same length as `dimensions`
        """
        # Transform
        for j in range(self.n_dims):
            if isinstance(transform, list):
                self.dimensions[j].set_transformer(transform[j])
            else:
                self.dimensions[j].set_transformer(transform)

    def set_transformer_by_type(self, transform, dim_type):
        """Sets the transformer of `dim_type` objects to `transform`

        Parameters
        ----------
        transform : str
           Sets all transformer of type `dim_type` to `transform`
        dim_type : type
            Can be `skopt.space.Real`, `skopt.space.Integer` or
             `skopt.space.Categorical`
        """
        # Transform
        for j in range(self.n_dims):
            if isinstance(self.dimensions[j], dim_type):
                self.dimensions[j].set_transformer(transform)

    def get_transformer(self):
        """Returns all transformers as list"""
        return [self.dimensions[j].transform_ for j in range(self.n_dims)]

    def transform(self, X):
        """Transform samples from the original space into a warped space.

        Note: this transformation is expected to be used to project samples
              into a suitable space for numerical optimization.

        Parameters
        ----------
        X : list of lists, shape=(n_samples, n_dims)
            The samples to transform.

        Returns
        -------
        Xt : array of floats, shape=(n_samples, transformed_n_dims)
            The transformed samples.
        """
        # Pack by dimension
        columns = []
        for dim in self.dimensions:
            columns.append([])

        for i in range(len(X)):
            for j in range(self.n_dims):
                columns[j].append(X[i][j])

        # Transform
        for j in range(self.n_dims):
            columns[j] = self.dimensions[j].transform(columns[j])

        # Repack as an array
        Xt = np.hstack([np.asarray(c).reshape((len(X), -1)) for c in columns])

        return Xt

    def inverse_transform(self, Xt):
        """Inverse transform samples from the warped space back to the
           original space.

        Parameters
        ----------
        Xt : array of floats, shape=(n_samples, transformed_n_dims)
            The samples to inverse transform.

        Returns
        -------
        X : list of lists, shape=(n_samples, n_dims)
            The original samples.
        """
        # Inverse transform
        columns = []
        start = 0
        Xt = np.asarray(Xt)
        for j in range(self.n_dims):
            dim = self.dimensions[j]
            offset = dim.transformed_size

            if offset == 1:
                columns.append(dim.inverse_transform(Xt[:, start]))
            else:
                columns.append(
                    dim.inverse_transform(Xt[:, start:start + offset]))

            start += offset

        # Transpose
        return _transpose_list_array(columns)

    @property
    def n_dims(self):
        """The dimensionality of the original space."""
        return len(self.dimensions)

    @property
    def transformed_n_dims(self):
        """The dimensionality of the warped space."""
        return sum([dim.transformed_size for dim in self.dimensions])

    @property
    def bounds(self):
        """The dimension bounds, in the original space."""
        b = []

        for dim in self.dimensions:
            if dim.size == 1:
                b.append(dim.bounds)
            else:
                b.extend(dim.bounds)

        return b

    def __contains__(self, point):
        """Check that `point` is within the bounds of the space."""
        for component, dim in zip(point, self.dimensions):
            if component not in dim:
                return False
        return True

    def __getitem__(self, dimension_names):
        """
        Lookup and return the search-space dimension with the given name.

        This allows for dict-like lookup of dimensions, for example:
        `space['foo']` returns the dimension named 'foo' if it exists,
        otherwise `None` is returned.

        It also allows for lookup of a list of dimension-names, for example:
        `space[['foo', 'bar']]` returns the two dimensions named
        'foo' and 'bar' if they exist.

        Parameters
        ----------
        dimension_names : str or list(str)
            Name of a single search-space dimension (str).
            List of names for search-space dimensions (list(str)).

        Returns
        -------
        dims tuple (index, Dimension), list(tuple(index, Dimension)), \
                (None, None)
            A single search-space dimension with the given name,
            or a list of search-space dimensions with the given names.
        """

        def _get(dimension_name):
            """Helper-function for getting a single dimension."""
            index = 0
            # Get the index of the search-space dimension using its name.
            for dim in self.dimensions:
                if dimension_name == dim.name:
                    return (index, dim)
                elif dimension_name == index:
                    return (index, dim)
                index += 1
            return (None, None)

        if isinstance(dimension_names, (str, int)):
            # Get a single search-space dimension.
            dims = _get(dimension_name=dimension_names)
        elif isinstance(dimension_names, (list, tuple)):
            # Get a list of search-space dimensions.
            # Note that we do not check whether the names are really strings.
            dims = [_get(dimension_name=name) for name in dimension_names]
        else:
            msg = "Dimension name should be either string or" \
                  "list of strings, but got {}."
            raise ValueError(msg.format(type(dimension_names)))

        return dims

    @property
    def transformed_bounds(self):
        """The dimension bounds, in the warped space."""
        b = []

        for dim in self.dimensions:
            if dim.transformed_size == 1:
                b.append(dim.transformed_bounds)
            else:
                b.extend(dim.transformed_bounds)

        return b

    @property
    def is_categorical(self):
        """Space contains exclusively categorical dimensions"""
        return all([isinstance(dim, Categorical) for dim in self.dimensions])

    @property
    def is_partly_categorical(self):
        """Space contains any categorical dimensions"""
        return any([isinstance(dim, Categorical) for dim in self.dimensions])

    @property
    def n_constant_dimensions(self):
        """Returns the number of constant dimensions which have zero degree of
        freedom, e.g. an Integer dimensions with (0., 0.) as bounds.
        """
        n = 0
        for dim in self.dimensions:
            if dim.is_constant:
                n += 1
        return n

    def distance(self, point_a, point_b):
        """Compute distance between two points in this space.

        Parameters
        ----------
        point_a : array
            First point.

        point_b : array
            Second point.
        """
        distance = 0.
        for a, b, dim in zip(point_a, point_b, self.dimensions):
            distance += dim.distance(a, b)

        return distance




def order_tree_model_dict(tree_model_dict):
    """
    Sorts a tree model dict provided by LightGBM from .json file to make it 
    suitable for use in ENTMOOT.

    Key name `tree_info` is specific to LightGBM and needs to be changed for
    other tree training libraries that support model dump to .json format.

    Parameters
    ----------
    tree_model_dict : dict,
        Includes all information on trees trained in LightGBM

    Returns
    -------
    ordered_tree_list : list,
        Ordered tree list compatible with `GbmModel`
    """
    tree_list = tree_model_dict['tree_info']
    ordered_tree_list = order_tree_list(tree_list)
    return ordered_tree_list

def order_tree_list(tree_list):
    """
    Sorts a tree list provided by LightGBM from .json file to make it 
    suitable for use in ENTMOOT.

    Parameters
    ----------
    tree_list : list,
        Unsorted list of trees from LightGBM

    Returns
    -------
    ordered_tree_list : list,
        Ordered tree list compatible with `GbmModel`
    """
    ordered_tree_list = []
    for tree in tree_list:
        ordered_node_list = order_node_list(tree)
        if ordered_node_list:
            ordered_tree_list.append(ordered_node_list)
    return ordered_tree_list

def order_node_list(tree):
    """
    Sorts a list of node dict from a LightGBM instance. Key `tree_structure` is 
    specific for LightGBM.

    Parameters
    ----------
    tree : list,
        Unsorted list of node dicts

    Returns
    -------
    ordered_node_list : list,
        Ordered list of node dicts compatible with `GbmModel`
    """
    node = []
    node.append(tree['tree_structure'])
    ordered_node_list = []
    add_next_nodes(ordered_node_list, node)
    return ordered_node_list
    
def add_next_nodes(ordered_node_list, node):
    """
    Processes LightGBM node and adds sorted node to `ordered_node_list`.

    Parameters
    ----------
    ordered_node_list : list,
        List of nodes to which sorted node is added

    node : list,
        List of unsorted LightGBM nodes for one tree

    Returns
    -------
    -
    """
    if node:   
        new_node = {}
        
        try:
            new_node['split_var'] = node[-1]['split_feature']
            temp_node_val = node[-1]['threshold']
            new_node['split_code_pred'] = round(temp_node_val,9)
        except KeyError:
            new_node['split_var'] = -1
            temp_node_val = node[-1]['leaf_value']
            new_node['split_code_pred'] = round(temp_node_val,9)

        ordered_node_list.append(new_node)

        temp_node = node.pop(-1)
        try:
            node.append(temp_node['right_child'])
        except KeyError:
            pass

        try:
            node.append(temp_node['left_child'])
        except KeyError:
            pass

        add_next_nodes(ordered_node_list, node)
    
    



"""
Copyright (c) 2016-2020 The scikit-optimize developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# -*- coding: utf-8 -*-
""" Inspired by https://github.com/jonathf/chaospy/blob/master/chaospy/
distributions/sampler/sequences/hammersley.py
"""
import numpy as np
from entmoot.sampler.halton import Halton
from entmoot.sampler.base import InitialPointGenerator
from entmoot.space.space import Space
from sklearn.utils import check_random_state


class Hammersly(InitialPointGenerator):
    """Creates `Hammersley` sequence samples.
    The Hammersley set is equivalent to the Halton sequence, except for one
    dimension is replaced with a regular grid. It is not recommended to
    generate a Hammersley sequence with more than 10 dimension.

    For ``dim == 1`` the sequence falls back to Van Der Corput sequence.

    References
    ----------
    T-T. Wong, W-S. Luk, and P-A. Heng, "Sampling with Hammersley and Halton
    Points," Journal of Graphics Tools, vol. 2, no. 2, 1997, pp. 9 - 24.

    Parameters
    ----------
    min_skip : int, default=-1
        minimum skipped seed number. When `min_skip != max_skip` and
        both are > -1, a random number is picked.
    max_skip : int, default=-1
        maximum skipped seed number. When `min_skip != max_skip` and
        both are > -1, a random number is picked.
    primes : tuple, default=None
        The (non-)prime base to calculate values along each axis. If
        empty, growing prime values starting from 2 will be used.
    """
    def __init__(self, min_skip=-1, max_skip=-1, primes=None):
        self.primes = primes
        self.min_skip = min_skip
        self.max_skip = max_skip

    def generate(self, dimensions, n_samples, random_state=None):
        """Creates samples from Hammersly set.

        Parameters
        ----------
        dimensions : list, shape (n_dims,)
            List of search space dimensions.
            Each search dimension can be defined either as

            - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
              dimensions),
            - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
              dimensions),
            - as a list of categories (for `Categorical` dimensions), or
            - an instance of a `Dimension` object (`Real`, `Integer` or
              `Categorical`).
        n_samples : int
            The order of the Hammersley sequence.
            Defines the number of samples.
        random_state : int, RandomState instance, or None (default)
            Set random state to something other than None for reproducible
            results.

        Returns
        -------
        np.array, shape=(n_dim, n_samples)
            Hammersley set
        """
        rng = check_random_state(random_state)
        halton = Halton(min_skip=self.min_skip, max_skip=self.max_skip,
                        primes=self.primes)
        space = Space(dimensions)
        n_dim = space.n_dims
        transformer = space.get_transformer()
        space.set_transformer("normalize")
        if n_dim == 1:
            out = halton.generate(dimensions, n_samples,
                                  random_state=rng)
        else:
            out = np.empty((n_dim, n_samples), dtype=float)
            out[:n_dim - 1] = np.array(halton.generate(
                [(0., 1.), ] * (n_dim - 1), n_samples,
                random_state=rng)).T

            out[n_dim - 1] = np.linspace(0, 1, n_samples + 2)[1:-1]
            out = space.inverse_transform(out.T)
        space.set_transformer(transformer)
        return out



"""
Copyright (c) 2016-2020 The scikit-optimize developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

NOTE: Changes were made to the scikit-optimize source code included here. 
For the most recent version of scikit-optimize we refer to:
https://github.com/scikit-optimize/scikit-optimize/

Copyright (c) 2019-2020 Alexander Thebelt.
"""

import numpy as np
from scipy.optimize import OptimizeResult
from joblib import dump as dump_
from joblib import load as load_

def is_supported(base_estimator):
    from entmoot.learning.tree_model import EntingRegressor
    return isinstance(base_estimator, (EntingRegressor, str, type(None)))


def cook_estimator(base_estimator, std_estimator=None, space=None, random_state=None, 
        base_estimator_params=None):
    """Cook a default estimator.

    For the special base_estimator called "DUMMY" the return value is None.
    This corresponds to sampling points at random, hence there is no need
    for an estimator.

    Parameters
    ----------
    base_estimator : "GBRT", creates LightGBM tree model based on base_estimator 

    std_estimator : DistandBasedStd instance, 
        Estimates model uncertainty of base_estimator

    space : Space instance

    random_state : int, RandomState instance, or None (default)
        Set random state to something other than None for reproducible
        results.

    base_estimator_params : dict
        Extra parameters provided to the base_estimator at init time.
    """

    from lightgbm import LGBMRegressor
    from entmoot.learning.tree_model import EntingRegressor

    if isinstance(base_estimator, str):
        base_estimator = base_estimator.upper()
        if base_estimator not in ["GBRT", "RF", "DUMMY"]:
            raise ValueError("Valid strings for the base_estimator parameter "
                             " are: 'GBRT', 'RF', or 'DUMMY' not "
                             "%s." % base_estimator)
    elif is_supported(base_estimator):
        base_estimator = \
            EntingRegressor(
                base_estimator=base_estimator,
                random_state=random_state
            )
    else:
        raise ValueError("base_estimator is not supported.")

    if base_estimator == "GBRT":
        gbrt = LGBMRegressor(boosting_type='gbdt',
                            objective='regression',
                            verbose=-1,
                            )
        base_estimator = EntingRegressor(base_estimator=gbrt,
                                        std_estimator=std_estimator,
                                        random_state=random_state)
    elif base_estimator == "RF":
        rf = LGBMRegressor(boosting_type='random_forest',
                            objective='regression',
                            verbose=0,
                            subsample_freq=1,
                            subsample=0.9,
                            bagging_seed= random_state
                            )
        base_estimator = EntingRegressor(base_estimator=rf,
                                        std_estimator=std_estimator,
                                        random_state=random_state)

    elif base_estimator == "DUMMY":
        return None

    if base_estimator_params is not None:
        base_estimator.set_params(**base_estimator_params)
        
    return base_estimator

def cook_std_estimator(std_estimator, 
                    space=None, 
                    random_state=None, 
                    std_estimator_params=None):
    """Cook a default uncertainty estimator.

    Parameters
    ----------
    std_estimator : string.
        A model is used to estimate uncertainty of `base_estimator`.
        Different types can be classified as exploration measures, i.e. move
        as far as possible from reference points, and penalty measures, i.e.
        stay as close as possible to reference points. Within these types, the 
        following uncertainty estimators are available:
        
        - exploration:
            - "BDD" for bounded-data distance, which uses squared euclidean
              distance to standardized data points
            - "L1BDD" for bounded-data distance, which uses manhattan
              distance to standardized data points
        
        - penalty:
            - "DDP" for data distance, which uses squared euclidean
              distance to standardized data points
            - "L1DDP" for data distance, which uses manhattan
              distance to standardized data points

    space : Space instance

    random_state : int, RandomState instance, or None (default)
        Set random state to something other than None for reproducible
        results.

    std_estimator_params : dict
        Extra parameters provided to the std_estimator at init time.
    """
    from entmoot.learning.distance_based_std import \
        DistanceBasedExploration, DistanceBasedPenalty

    if std_estimator == "BDD":
        std_estimator = \
            DistanceBasedExploration(
                metric="sq_euclidean",
            )

    elif std_estimator == "DDP":
        std_estimator = \
            DistanceBasedPenalty(
                metric="sq_euclidean",
            )

    elif std_estimator == "L1BDD":
        std_estimator = \
            DistanceBasedExploration(
                metric="manhattan",
            )

    elif std_estimator == "L1DDP":
        std_estimator = \
            DistanceBasedPenalty(
                metric="manhattan",
            )

    std_estimator.set_params(**std_estimator_params)
    return std_estimator

def cook_initial_point_generator(generator, **kwargs):
    """Cook a default initial point generator.

    For the special generator called "random" the return value is None.

    Parameters
    ----------
    generator : "lhs", "sobol", "halton", "hammersly", "grid", "random" \
            or InitialPointGenerator instance"
        Should inherit from `skopt.sampler.InitialPointGenerator`.

    kwargs : dict
        Extra parameters provided to the generator at init time.
    """
    from entmoot.sampler import Sobol, Lhs, Hammersly, Halton, Grid
    from entmoot.sampler import InitialPointGenerator

    if generator is None:
        generator = "random"
    elif isinstance(generator, str):
        generator = generator.lower()
        if generator not in ["sobol", "halton", "hammersly", "lhs", "random",
                             "grid"]:
            raise ValueError("Valid strings for the generator parameter "
                             " are: 'sobol', 'lhs', 'halton', 'hammersly',"
                             "'random', or 'grid' not "
                             "%s." % generator)
    elif not isinstance(generator, InitialPointGenerator):
        raise ValueError("generator has to be an InitialPointGenerator."
                         "Got %s" % (str(type(generator))))

    if isinstance(generator, str):
        if generator == "sobol":
            generator = Sobol()
        elif generator == "halton":
            generator = Halton()
        elif generator == "hammersly":
            generator = Hammersly()
        elif generator == "lhs":
            generator = Lhs()
        elif generator == "grid":
            generator = Grid()
        elif generator == "random":
            return None
    generator.set_params(**kwargs)
    return generator

def is_listlike(x):
    return isinstance(x, (list, tuple))


def is_2Dlistlike(x):
    return np.all([is_listlike(xi) for xi in x])


def check_x_in_space(x, space):
    if is_2Dlistlike(x):
        if not np.all([p in space for p in x]):
            raise ValueError("Not all points are within the bounds of"
                             " the space.")
        if any([len(p) != len(space.dimensions) for p in x]):
            raise ValueError("Not all points have the same dimensions as"
                             " the space.")
    elif is_listlike(x):
        if x not in space:
            raise ValueError("Point (%s) is not within the bounds of"
                             " the space (%s)."
                             % (x, space.bounds))
        if len(x) != len(space.dimensions):
            raise ValueError("Dimensions of point (%s) and space (%s) do not match"
                             % (x, space.bounds))

def create_result(Xi, yi, space=None, rng=None, specs=None, models=None,
                    model_mu=None, model_std=None, gurobi_mipgap=None):
    """
    Initialize an `OptimizeResult` object.

    Parameters
    ----------
    Xi : list of lists, shape (n_iters, n_features)
        Location of the minimum at every iteration.

    yi : array-like, shape (n_iters,)
        Minimum value obtained at every iteration.

    space : Space instance, optional
        Search space.

    rng : RandomState instance, optional
        State of the random state.

    specs : dict, optional
        Call specifications.

    models : list, optional
        List of fit surrogate models.

    Returns
    -------
    res : `OptimizeResult`, scipy object
        OptimizeResult instance with the required information.
    """
    res = OptimizeResult()
    yi = np.asarray(yi)
    if np.ndim(yi) == 2:
        res.log_time = np.ravel(yi[:, 1])
        yi = np.ravel(yi[:, 0])
    best = np.argmin(yi)
    res.x = Xi[best]
    res.fun = yi[best]
    res.func_vals = yi
    res.x_iters = Xi
    res.models = models
    res.model_mu = model_mu
    res.model_std = model_std
    res.gurobi_mipgap = gurobi_mipgap
    res.space = space
    res.random_state = rng
    res.specs = specs
    return res

def dump(res, filename, store_objective=True, **kwargs):
    """
    Store an skopt optimization result into a file.

    Parameters
    ----------
    res : `OptimizeResult`, scipy object
        Optimization result object to be stored.

    filename : string or `pathlib.Path`
        The path of the file in which it is to be stored. The compression
        method corresponding to one of the supported filename extensions ('.z',
        '.gz', '.bz2', '.xz' or '.lzma') will be used automatically.

    store_objective : boolean, default=True
        Whether the objective function should be stored. Set `store_objective`
        to `False` if your objective function (`.specs['args']['func']`) is
        unserializable (i.e. if an exception is raised when trying to serialize
        the optimization result).

        Notice that if `store_objective` is set to `False`, a deep copy of the
        optimization result is created, potentially leading to performance
        problems if `res` is very large. If the objective function is not
        critical, one can delete it before calling `skopt.dump()` and thus
        avoid deep copying of `res`.

    **kwargs : other keyword arguments
        All other keyword arguments will be passed to `joblib.dump`.
    """
    if store_objective:
        dump_(res, filename, **kwargs)

    elif 'func' in res.specs['args']:
        # If the user does not want to store the objective and it is indeed
        # present in the provided object, then create a deep copy of it and
        # remove the objective function before dumping it with joblib.dump.
        res_without_func = deepcopy(res)
        del res_without_func.specs['args']['func']
        dump_(res_without_func, filename, **kwargs)

    else:
        # If the user does not want to store the objective and it is already
        # missing in the provided object, dump it without copying.
        dump_(res, filename, **kwargs)


def load(filename, **kwargs):
    """
    Reconstruct a skopt optimization result from a file
    persisted with skopt.dump.

    .. note::
        Notice that the loaded optimization result can be missing
        the objective function (`.specs['args']['func']`) if `skopt.dump`
        was called with `store_objective=False`.

    Parameters
    ----------
    filename : string or `pathlib.Path`
        The path of the file from which to load the optimization result.

    **kwargs : other keyword arguments
        All other keyword arguments will be passed to `joblib.load`.

    Returns
    -------
    res : `OptimizeResult`, scipy object
        Reconstructed OptimizeResult instance.
    """
    return load_(filename, **kwargs)


"""
Copyright (c) 2016-2020 The scikit-optimize developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division
import numpy as np
from sklearn.preprocessing import LabelBinarizer
from sklearn.utils import column_or_1d


class Transformer(object):
    """Base class for all 1-D transformers.
    """
    def fit(self, X):
        return self

    def transform(self, X):
        raise NotImplementedError

    def inverse_transform(self, X):
        raise NotImplementedError


class Identity(Transformer):
    """Identity transform.
    """

    def transform(self, X):
        return X

    def inverse_transform(self, Xt):
        return Xt


class StringEncoder(Transformer):
    """StringEncoder transform.
       The transform will cast everything to a
       string and the inverse transform will cast to the type defined in dtype.
    """

    def __init__(self, dtype=str):
        super(StringEncoder, self).__init__()
        self.dtype = dtype

    def fit(self, X):
        """Fit a list or array of categories. All elements must be from the
        same type.

        Parameters
        ----------
        X : array-like, shape=(n_categories,)
            List of categories.
        """
        if len(X) > 0:
            self.dtype = type(X[0])

    def transform(self, X):
        """Transform an array of categories to a string encoded representation.

        Parameters
        ----------
        X : array-like, shape=(n_samples,)
            List of categories.

        Returns
        -------
        Xt : array-like, shape=(n_samples,)
            The string encoded categories.
        """
        return [str(x) for x in X]

    def inverse_transform(self, Xt):
        """Inverse transform string encoded categories back to their original
           representation.

        Parameters
        ----------
        Xt : array-like, shape=(n_samples,)
            String encoded categories.

        Returns
        -------
        X : array-like, shape=(n_samples,)
            The original categories.
        """
        return [self.dtype(x) for x in Xt]


class LogN(Transformer):
    """Base N logarithm transform."""

    def __init__(self, base):
        self._base = base

    def transform(self, X):
        return np.log10(np.asarray(X, dtype=np.float)) / np.log10(self._base)

    def inverse_transform(self, Xt):
        return self._base ** np.asarray(Xt, dtype=np.float)


class CategoricalEncoder(Transformer):
    """OneHotEncoder that can handle categorical variables."""

    def __init__(self):
        """Convert labeled categories into one-hot encoded features."""
        self._lb = LabelBinarizer()

    def fit(self, X):
        """Fit a list or array of categories.

        Parameters
        ----------
        X : array-like, shape=(n_categories,)
            List of categories.
        """
        self.mapping_ = {v: i for i, v in enumerate(X)}
        self.inverse_mapping_ = {i: v for v, i in self.mapping_.items()}
        self._lb.fit([self.mapping_[v] for v in X])
        self.n_classes = len(self._lb.classes_)

        return self

    def transform(self, X):
        """Transform an array of categories to a one-hot encoded representation.

        Parameters
        ----------
        X : array-like, shape=(n_samples,)
            List of categories.

        Returns
        -------
        Xt : array-like, shape=(n_samples, n_categories)
            The one-hot encoded categories.
        """
        return self._lb.transform([self.mapping_[v] for v in X])

    def inverse_transform(self, Xt):
        """Inverse transform one-hot encoded categories back to their original
           representation.

        Parameters
        ----------
        Xt : array-like, shape=(n_samples, n_categories)
            One-hot encoded categories.

        Returns
        -------
        X : array-like, shape=(n_samples,)
            The original categories.
        """
        Xt = np.asarray(Xt)
        return [
            self.inverse_mapping_[i] for i in self._lb.inverse_transform(Xt)
        ]


class LabelEncoder(Transformer):
    """LabelEncoder that can handle categorical variables."""
    def __init__(self, X=None):
        if X is not None:
            self.fit(X)

    def fit(self, X):
        """Fit a list or array of categories.

        Parameters
        ----------
        X : array-like, shape=(n_categories,)
            List of categories.
        """
        X = np.asarray(X)
        if X.dtype == object:
            self.mapping_ = {v: i for i, v in enumerate(X)}
        else:
            i = 0
            self.mapping_ = {}
            for v in np.unique(X):
                self.mapping_[v] = i
                i += 1
        self.inverse_mapping_ = {i: v for v, i in self.mapping_.items()}
        return self

    def transform(self, X):
        """Transform an array of categories to a one-hot encoded
        representation.

        Parameters
        ----------
        X : array-like, shape=(n_samples,)
            List of categories.

        Returns
        -------
        Xt : array-like, shape=(n_samples, n_categories)
            The integer categories.
        """
        X = np.asarray(X)
        return [self.mapping_[v] for v in X]

    def inverse_transform(self, Xt):
        """Inverse transform integer categories back to their original
           representation.

        Parameters
        ----------
        Xt : array-like, shape=(n_samples, n_categories)
            Integer categories.

        Returns
        -------
        X : array-like, shape=(n_samples,)
            The original categories.
        """
        if isinstance(Xt, (float, np.float64)):
            Xt = [Xt]
        else:
            Xt = np.asarray(Xt)
        return [
            self.inverse_mapping_[int(np.round(i))] for i in Xt
        ]


class Normalize(Transformer):
    """
    Scales each dimension into the interval [0, 1].

    Parameters
    ----------
    low : float
        Lower bound.

    high : float
        Higher bound.

    is_int : bool, default=True
        Round and cast the return value of `inverse_transform` to integer. Set
        to `True` when applying this transform to integers.
    """
    def __init__(self, low, high, is_int=False):
        self.low = float(low)
        self.high = float(high)
        self.is_int = is_int
        self._eps = 1e-8

    def transform(self, X):
        X = np.asarray(X)
        if self.is_int:
            if np.any(np.round(X) > self.high):
                raise ValueError("All integer values should"
                                 "be less than %f" % self.high)
            if np.any(np.round(X) < self.low):
                raise ValueError("All integer values should"
                                 "be greater than %f" % self.low)
        else:
            if np.any(X > self.high + self._eps):
                raise ValueError("All values should"
                                 "be less than %f" % self.high)
            if np.any(X < self.low - self._eps):
                raise ValueError("All values should"
                                 "be greater than %f" % self.low)
        if (self.high - self.low) == 0.:
            return X * 0.
        if self.is_int:
            return (np.round(X).astype(np.int) - self.low) /\
                   (self.high - self.low)
        else:
            return (X - self.low) / (self.high - self.low)

    def inverse_transform(self, X):
        X = np.asarray(X)
        if np.any(X > 1.0 + self._eps):
            raise ValueError("All values should be less than 1.0")
        if np.any(X < 0.0 - self._eps):
            raise ValueError("All values should be greater than 0.0")
        X_orig = X * (self.high - self.low) + self.low
        if self.is_int:
            return np.round(X_orig).astype(np.int)
        return X_orig


class Pipeline(Transformer):
    """
    A lightweight pipeline to chain transformers.

    Parameters
    ----------
    transformers : list
        A list of Transformer instances.
    """
    def __init__(self, transformers):
        self.transformers = list(transformers)
        for transformer in self.transformers:
            if not isinstance(transformer, Transformer):
                raise ValueError(
                    "Provided transformers should be a Transformer "
                    "instance. Got %s" % transformer
                )

    def fit(self, X):
        for transformer in self.transformers:
            transformer.fit(X)
        return self

    def transform(self, X):
        for transformer in self.transformers:
            X = transformer.transform(X)
        return X

    def inverse_transform(self, X):
        for transformer in self.transformers[::-1]:
            X = transformer.inverse_transform(X)
        return X



import collections as coll
import sys

class GbmModel:
    """Define a gbm model.

    A `GbmModel` is a object-oriented class to enocode tree-based models from
    different libraries. Supported tree training libraries are: LightGBM.

    Use this class as an interface to read in model data from tree model
    training libraries and translate it to the ENTMOOT native structure.

    Parameters
    ----------
    tree_list : list
        Ordered tree list that obeys ENTMOOT input specifications

    Attributes
    ----------
    -

    """
    def __init__(self, tree_list):
        self.load_ordered_tree_dict(tree_list)

    def _build_tree(self, tree):
        """Define `GbmNode` based on single tree.

        Parameters
        ----------
        trees : list of dicts,
            Specifies list of node dicts
        n_trees : number of trees in `tree_list`

        Returns
        -------
        gbm_node : GbmNode
        """
        node = tree.pop(0)
        split_var = node['split_var']
        split_code_pred = node['split_code_pred']

        return GbmNode(
            split_var=split_var,
            split_code_pred=split_code_pred,
            tree=tree
        )

    def load_ordered_tree_dict(self, tree_list):
        """Define attributes used in `GbmModel`.

        Parameters
        ----------
        tree_list : list of trees,
            Specifies list of trees which contain list of node dicts

        Returns
        -------
        -
        """
        self.n_trees = len(tree_list)
        self.trees = [self._build_tree(tree) for tree in tree_list]

    def get_leaf_encodings(self, tree):
        yield from self.trees[tree].get_leaf_encodings()

    def get_branch_encodings(self, tree):
        yield from self.trees[tree].get_branch_encodings()

    def get_leaf_weight(self, tree, encoding):
        return self.trees[tree].get_leaf_weight(encoding)

    def get_leaf_weights(self, tree):
        return self.trees[tree].get_leaf_weights()

    def get_branch_partition_pair(self, tree, encoding):
        return self.trees[tree].get_branch_partition_pair(encoding)

    def get_left_leaves(self, tree, encoding):
        yield from (
            encoding + s
            for s in self.trees[tree].get_left_leaves(encoding)
        )

    def get_right_leaves(self, tree, encoding):
        yield from (
            encoding + s
            for s in self.trees[tree].get_right_leaves(encoding)
        )

    def get_branch_partition_pairs(self, tree, leaf_encoding):
        yield from self.trees[tree].get_branch_partition_pairs(leaf_encoding)

    def get_participating_variables(self, tree, leaf):
        return set(self.trees[tree].get_participating_variables(leaf))

    def get_all_participating_variables(self, tree):
        return set(self.trees[tree].get_all_participating_variables())

    def get_var_lower(self, tree, leaf, var, lower):
        return self.trees[tree].get_var_lower(leaf, var, lower)

    def get_var_upper(self, tree, leaf, var, upper):
        return self.trees[tree].get_var_upper(leaf, var, upper)

    def get_var_interval(self, tree, leaf, var, var_interval):
        return self.trees[tree].get_var_interval(leaf, var, var_interval)

    def get_var_break_points(self):
        var_breakpoints = {}
        for tree in self.trees:
            for var, breakpoint in tree.get_all_partition_pairs():
                try:
                    var_breakpoints[var].add(breakpoint)
                except KeyError:
                    var_breakpoints[var] = set([breakpoint])

        for k in var_breakpoints.keys():
            var_breakpoints[k] = sorted(var_breakpoints[k])

        return var_breakpoints

    def get_leaf_count(self):
        resulting_counter = sum(
            (tree.get_leaf_count() for tree in self.trees),
            coll.Counter()
        )
        del resulting_counter['leaf']
        return resulting_counter


class GbmType:
    pass


class GbmNode(GbmType):
    """Defines a gbm node which can be a split or leaf.

    Initializing `GbmNode` triggers recursive initialization of all nodes that 
    appear in a tree

    Parameters
    ----------
    split_var : int
        Index of split variable used
    split_code_pred : list
        Value that defines split
    tree : list
        List of node dicts that define the tree

    Attributes
    ----------
    split_var : int
        Index of split variable used
    split_code_pred : list
        Value that defines split
    tree : list
        List of node dicts that define the tree
    """
    def __init__(
        self,
        split_var,
        split_code_pred,
        tree
    ):
        self.split_var = split_var
        self.split_code_pred = split_code_pred

        # check if tree is empty
        if not tree:
            print("EmptyTreeModelError: LightGBM was not able to train a "
            "tree model based on your parameter specifications. This can "
            "usually be fixed by increasing the number of `n_initial_points` or "
            "reducing `min_child_samples` via `base_estimator_kwargs` (default "
            "is 20). Alternatively, change `acq_optimizer='sampling'`.")
            import sys
            sys.exit(1)

        # read left node 
        node = tree.pop(0)
        split_var = node['split_var']
        split_code_pred = node['split_code_pred']

        # split_var value of -1 refers to leaf node
        if split_var == -1:
            self.left = LeafNode(
                split_code_pred = split_code_pred
            )
        else:
            self.left = GbmNode(
                split_var = split_var,
                split_code_pred = split_code_pred,
                tree = tree
            )

        # read right node
        node = tree.pop(0)
        split_var = node['split_var']
        split_code_pred = node['split_code_pred']

        if split_var == -1:
            self.right = LeafNode(
                split_code_pred=split_code_pred
            )
        else:
            self.right = GbmNode(
                split_var=split_var,
                split_code_pred=split_code_pred,
                tree=tree
            )

    def __repr__(self):
        return ', '.join([
            str(x)
            for x in [
                self.split_var,
                self.split_code_pred]
        ])

    def _get_next_node(self, direction):
        return self.right if int(direction) else self.left

    def get_leaf_encodings(self, current_string=''):
        yield from self.left.get_leaf_encodings(current_string+'0')
        yield from self.right.get_leaf_encodings(current_string+'1')

    def get_branch_encodings(self, current_string=''):
        yield current_string
        yield from self.left.get_branch_encodings(current_string+'0')
        yield from self.right.get_branch_encodings(current_string+'1')

    def get_leaf_weight(self, encoding):
        next_node = self.right if int(encoding[0]) else self.left
        return next_node.get_leaf_weight(encoding[1:])

    def get_leaf_weights(self):
        yield from self.left.get_leaf_weights()
        yield from self.right.get_leaf_weights()

    def get_branch_partition_pair(self, encoding):
        if not encoding:
            return self.split_var, self.split_code_pred
        else:
            next_node = self._get_next_node(encoding[0])
            return next_node.get_branch_partition_pair(encoding[1:])

    def get_left_leaves(self, encoding):
        if encoding:
            next_node = self._get_next_node(encoding[0])
            yield from next_node.get_left_leaves(encoding[1:])
        else:
            yield from self.left.get_leaf_encodings('0')

    def get_right_leaves(self, encoding):
        if encoding:
            next_node = self._get_next_node(encoding[0])
            yield from next_node.get_right_leaves(encoding[1:])
        else:
            yield from self.right.get_leaf_encodings('1')

    def get_branch_partition_pairs(self, encoding):
        yield (self.split_var, self.split_code_pred)
        try:
            next_node = self.right if int(encoding[0]) else self.left
            next_gen = next_node.get_branch_partition_pairs(encoding[1:])
        except IndexError:
            next_gen = []
        yield from next_gen

    def get_participating_variables(self, encoding):
        yield self.split_var
        next_node = self.right if int(encoding[0]) else self.left
        yield from next_node.get_participating_variables(encoding[1:])

    def get_all_participating_variables(self):
        yield self.split_var
        yield from self.left.get_all_participating_variables()
        yield from self.right.get_all_participating_variables()

    def get_all_partition_pairs(self):
        yield (self.split_var, self.split_code_pred)
        yield from self.left.get_all_partition_pairs()
        yield from self.right.get_all_partition_pairs()

    def get_var_lower(self, encoding, var, lower):
        if encoding:
            if self.split_var == var:
                if int(encoding[0]) == 1:
                    assert lower <= self.split_code_pred
                    lower = self.split_code_pred
            next_node = self.right if int(encoding[0]) else self.left
            return next_node.get_var_lower(encoding[1:], var, lower)
        else:
            return lower

    def get_var_upper(self, encoding, var, upper):
        if encoding:
            if self.split_var == var:
                if int(encoding[0]) == 0:
                    assert upper >= self.split_code_pred
                    upper = self.split_code_pred
            next_node = self.right if int(encoding[0]) else self.left
            return next_node.get_var_upper(encoding[1:], var, upper)
        else:
            return upper

    def get_var_interval(self, encoding, var, var_interval):
        if int(encoding[0]):
            next_node = self.right
            if self.split_var == var:
                var_interval = (self.split_code_pred, var_interval[1])
        else:
            next_node = self.left
            if self.split_var == var:
                var_interval = (var_interval[0], self.split_code_pred)
        return next_node.get_var_interval(encoding[1:], var, var_interval)

    def get_leaf_count(self):
        left_count = self.left.get_leaf_count()
        right_count = self.right.get_leaf_count()
        joint_count = left_count+right_count

        left_key = (self.split_var, self.split_code_pred, 0)
        right_key = (self.split_var, self.split_code_pred, 1)
        joint_count[left_key] += left_count['leaf']
        joint_count[right_key] += right_count['leaf']
        return joint_count

class LeafNode(GbmType):
    """Defines a child class of `GbmType`. Leaf nodes have `split_var = -1` and 
    `split_code_pred` as leaf value defined by training."""
    def __init__(self,
                 split_code_pred):
        self.split_code_pred = split_code_pred

    def __repr__(self):
        return ', '.join([
            str(x)
            for x in [
                'Leaf',
                self.split_code_pred]
        ])

    def switch_to_maximisation(self):
        """Changes the sign of tree model prediction by changing signs of 
        leaf values."""
        self.split_code_pred = -1*self.split_code_pred


    def get_leaf_encodings(self, current_string=''):
        yield current_string

    def get_branch_encodings(self, current_string=''):
        yield from []

    def get_leaf_weight(self, encoding):
        assert not encoding
        return self.split_code_pred

    def get_leaf_weights(self):
        yield self.split_code_pred

    def get_branch_partition_pair(self, encoding):
        raise Exception('Should not get here.')

    def get_left_leaves(self, encoding):
        raise Exception('Should not get here.')

    def get_right_leaves(self, encoding):
        raise Exception('Should not get here.')

    def get_branch_partition_pairs(self, encoding):
        assert not encoding
        yield from []

    def get_participating_variables(self, encoding):
        assert not encoding
        yield from []

    def get_all_participating_variables(self):
        yield from []

    def get_all_partition_pairs(self):
        yield from []

    def get_var_lower(self, encoding, var, lower):
        assert not encoding
        return lower

    def get_var_upper(self, encoding, var, upper):
        assert not encoding
        return upper

    def get_var_interval(self, encoding, var, var_interval):
        assert not encoding
        return var_interval

    def get_leaf_count(self):
        return coll.Counter({'leaf': 1})


"""
Copyright (c) 2016-2020 The scikit-optimize developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Inspired by https://github.com/jonathf/chaospy/blob/master/chaospy/
distributions/sampler/sequences/grid.py
"""
import numpy as np
from entmoot.sampler.base import InitialPointGenerator
from entmoot.space.space import Space
from sklearn.utils import check_random_state


def _quadrature_combine(args):
    args = [np.asarray(arg).reshape(len(arg), -1) for arg in args]
    shapes = [arg.shape for arg in args]

    size = np.prod(shapes, 0)[0] * np.sum(shapes, 0)[1]
    if size > 10 ** 9:
        raise MemoryError("Too large sets")

    out = args[0]
    for arg in args[1:]:
        out = np.hstack([
            np.tile(out, len(arg)).reshape(-1, out.shape[1]),
            np.tile(arg.T, len(out)).reshape(arg.shape[1], -1).T,
        ])
    return out


def _create_uniform_grid_exclude_border(n_dim, order):
    assert order > 0
    assert n_dim > 0
    x_data = np.arange(1, order + 1) / (order + 1.)
    x_data = _quadrature_combine([x_data] * n_dim)
    return x_data


def _create_uniform_grid_include_border(n_dim, order):
    assert order > 1
    assert n_dim > 0
    x_data = np.arange(0, order) / (order - 1.)
    x_data = _quadrature_combine([x_data] * n_dim)
    return x_data


def _create_uniform_grid_only_border(n_dim, order):
    assert n_dim > 0
    assert order > 1
    x = [[0., 1.]] * (n_dim - 1)
    x.append(list(np.arange(0, order) / (order - 1.)))
    x_data = _quadrature_combine(x)
    return x_data


class Grid(InitialPointGenerator):
    """Generate samples from a regular grid.

    Parameters
    ----------
    border : str, default='exclude'
        defines how the samples are generated:
        - 'include' : Includes the border into the grid layout
        - 'exclude' : Excludes the border from the grid layout
        - 'only' : Selects only points at the border of the dimension
    use_full_layout : boolean, default=True
        When True, a  full factorial design is generated and
        missing points are taken from the next larger full factorial
        design, depending on `append_border`
        When False, the next larger  full factorial design is
        generated and points are randomly selected from it.
    append_border : str, default="only"
        When use_full_layout is True, this parameter defines how the missing
        points will be generated from the next larger grid layout:
        - 'include' : Includes the border into the grid layout
        - 'exclude' : Excludes the border from the grid layout
        - 'only' : Selects only points at the border of the dimension
    """

    def __init__(self, border="exclude", use_full_layout=True,
                 append_border="only"):
        self.border = border
        self.use_full_layout = use_full_layout
        self.append_border = append_border

    def generate(self, dimensions, n_samples, random_state=None):
        """Creates samples from a regular grid.

        Parameters
        ----------
        dimensions : list, shape (n_dims,)
            List of search space dimensions.
            Each search dimension can be defined either as

            - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
              dimensions),
            - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
              dimensions),
            - as a list of categories (for `Categorical` dimensions), or
            - an instance of a `Dimension` object (`Real`, `Integer` or
              `Categorical`).

        n_samples : int
            The order of the Halton sequence. Defines the number of samples.
        random_state : int, RandomState instance, or None (default)
            Set random state to something other than None for reproducible
            results.

        Returns
        -------
        np.array, shape=(n_dim, n_samples)
            grid set
        """
        rng = check_random_state(random_state)
        space = Space(dimensions)
        n_dim = space.n_dims
        transformer = space.get_transformer()
        space.set_transformer("normalize")

        if self.border == "include":
            if self.use_full_layout:
                order = int(np.floor(np.sqrt(n_samples)))
            else:
                order = int(np.ceil(np.sqrt(n_samples)))
            if order < 2:
                order = 2
            h = _create_uniform_grid_include_border(n_dim, order)
        elif self.border == "exclude":
            if self.use_full_layout:
                order = int(np.floor(np.sqrt(n_samples)))
            else:
                order = int(np.ceil(np.sqrt(n_samples)))
            if order < 1:
                order = 1
            h = _create_uniform_grid_exclude_border(n_dim, order)
        elif self.border == "only":
            if self.use_full_layout:
                order = int(np.floor(n_samples / 2.))
            else:
                order = int(np.ceil(n_samples / 2.))
            if order < 2:
                order = 2
            h = _create_uniform_grid_exclude_border(n_dim, order)
        else:
            raise ValueError("Wrong value for border")
        if np.size(h, 0) > n_samples:
            rng.shuffle(h)
            h = h[:n_samples, :]
        elif np.size(h, 0) < n_samples:
            if self.append_border == "only":
                order = int(np.ceil((n_samples - np.size(h, 0)) / 2.))
                if order < 2:
                    order = 2
                h2 = _create_uniform_grid_only_border(n_dim, order)
            elif self.append_border == "include":
                order = int(np.ceil(np.sqrt(n_samples - np.size(h, 0))))
                if order < 2:
                    order = 2
                h2 = _create_uniform_grid_include_border(n_dim, order)
            elif self.append_border == "exclude":
                order = int(np.ceil(np.sqrt(n_samples - np.size(h, 0))))
                if order < 1:
                    order = 1
                h2 = _create_uniform_grid_exclude_border(n_dim, order)
            else:
                raise ValueError("Wrong value for append_border")
            h = np.vstack((h, h2[:(n_samples - np.size(h, 0))]))
            rng.shuffle(h)
        else:
            rng.shuffle(h)
        h = space.inverse_transform(h)
        space.set_transformer(transformer)
        return h



"""
Copyright (c) 2016-2020 The scikit-optimize developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

NOTE: Changes were made to the scikit-optimize source code included here. 
For the most recent version of scikit-optimize we refer to:
https://github.com/scikit-optimize/scikit-optimize/

Copyright (c) 2019-2020 Alexander Thebelt.
"""

from entmoot.optimizer.optimizer import Optimizer

import copy
import inspect
import numbers

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

def entmoot_minimize(
    func,
    dimensions,
    n_calls=60,
    batch_size=None,
    batch_strategy="cl_mean",
    n_points=10000,
    base_estimator="GBRT",
    std_estimator="BDD", 
    n_initial_points=50, 
    initial_point_generator="random",
    acq_func="LCB", 
    acq_optimizer="global",
    x0=None,
    y0=None,  
    random_state=None, 
    acq_func_kwargs=None, 
    acq_optimizer_kwargs=None,
    base_estimator_kwargs=None,
    std_estimator_kwargs=None,
    model_queue_size=None,
    verbose=False,
):
    """Run bayesian optimisation loop.

    An `Optimizer` is initialized which represents the steps of a bayesian 
    optimisation loop. To use it you need to provide your own loop mechanism. 

    Parameters specify how `Optimizer` is initialized, which function is 
    optimized and how each iteration is conducted.

    Parameters
    ----------
    func : BenchmarkFunction instance,
        Use class entmoot.benchmarks.BenchmarkFunction as a base to incorporate
        own black-box functions

    dimensions : list, shape (n_dims,)
        List of search space dimensions.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

        *NOTE: `Integer` and `Categorical` variables are not yet supported and
        will be added in the next release.

    n_calls : int, default: 60
        Budget of func calls that shall be used. `n_calls` includes initial
        points as well as suggested model-based points.

    batch_size : int, default: None
        Number of suggestions given per sequential iteration. For `None` only
        one suggestion per iteration is given.

    batch_strategy : string, default: "cl_mean"
        In case `batch_size` is not None, `batch_strategy` specifies the 
        strategy used to suggest multiple points per iteration. Available
        strategies are ["cl_min", "cl_mean", "cl_max"]

    n_points : int, default: 10,000
        For `acq_optimizer` = "sampling", `n_points` specifies the number of
        random points used to sample the acquisition function to derive its
        minimum.

    base_estimator : string, default: "GBRT",
        A default LightGBM surrogate model of the corresponding type is used  
        minimize `func`.
        The following model types are available:
        
        - "GBRT" for gradient-boosted trees

        - "RF" for random forests

        *NOTE: More model types will be added in the future

    std_estimator : string, default: "BDD",
        A model is used to estimate uncertainty of `base_estimator`.
        Different types can be classified as exploration measures, i.e. move
        as far as possible from reference points, and penalty measures, i.e.
        stay as close as possible to reference points. Within these types, the 
        following uncertainty estimators are available:
        
        - exploration:
            - "BDD" for bounded-data distance, which uses squared euclidean
              distance to standardized data points
            - "L1BDD" for bounded-data distance, which uses manhattan
              distance to standardized data points
        
        - penalty:
            - "DDP" for data distance, which uses squared euclidean
              distance to standardized data points
            - "L1DDP" for data distance, which uses manhattan
              distance to standardized data points

        *NOTE: More model uncertainty estimators will be added in the future

    n_initial_points : int, default: 50
        Number of evaluations of `func` with initialization points
        before approximating it with `base_estimator`. Initial point
        generator can be changed by setting `initial_point_generator`. For 
        `n_initial_points` <= 20 we need to set `min_child_samples` <= 20 in 
        `base_estimator_kwargs` so LightGBM can train tree models based on small
        data sets.

    initial_point_generator : str, InitialPointGenerator instance, \
            default: "random"
        Sets a initial points generator. Can be either

        - "random" for uniform random numbers,
        - "sobol" for a Sobol sequence,
        - "halton" for a Halton sequence,
        - "hammersly" for a Hammersly sequence,
        - "lhs" for a latin hypercube sequence,
        - "grid" for a uniform grid sequence

    acq_func : string, default: "LCB"
        Function to minimize over the posterior distribution. Can be either

        - "LCB" for lower confidence bound.

    acq_optimizer : string, default: "sampling"
        Method to minimize the acquisition function. The fit model
        is updated with the optimal value obtained by optimizing `acq_func`
        with `acq_optimizer`.

        - If set to "sampling", then `acq_func` is optimized by computing
          `acq_func` at `n_points` randomly sampled points.
        - If set to "global", then `acq_optimizer` is optimized by using
          global solver to find minimum of `acq_func`.

    x0 : list, list of lists or `None`
        Initial input points.
        - If it is a list of lists, use it as a list of input points.
        - If it is a list, use it as a single initial input point.
        - If it is `None`, no initial input points are used.

    y0 : list, scalar or `None`
        Evaluation of initial input points.
        - If it is a list, then it corresponds to evaluations of the function
          at each element of `x0` : the i-th element of `y0` corresponds
          to the function evaluated at the i-th element of `x0`.
        - If it is a scalar, then it corresponds to the evaluation of the
          function at `x0`.
        - If it is None and `x0` is provided, then the function is evaluated
          at each element of `x0`.

    random_state : int, RandomState instance, or None (default)
        Set random state to something other than None for reproducible
        results.

    acq_func_kwargs : dict
        Additional arguments to be passed to the acquisition function.

    acq_optimizer_kwargs : dict
        Additional arguments to be passed to the acquisition optimizer.

    base_estimator_kwargs : dict
        Additional arguments to be passed to the base_estimator.

    std_estimator_kwargs : dict
        Additional arguments to be passed to the std_estimator.

    model_queue_size : int or None, default: None
        Keeps list of models only as long as the argument given. In the
        case of None, the list has no capped length.

    verbose : bool or int:
        - If it is `True`, general solver information is printed at every 
          iteration
        - If it is `False`, no output is printed

        - If it is 0, same as if `False`
        - If it is 1, same as if `True`
        - If it is 2, general solver information is printed at every iteration
          as well as detailed solver information, i.e. gurobi log

    Returns
    -------
    res : `OptimizeResult`, scipy object
        OptimizeResult instance with the required information.
    """

    specs = {"args": copy.copy(inspect.currentframe().f_locals),
             "function": inspect.currentframe().f_code.co_name}

    acq_optimizer_kwargs = {
        "n_points": n_points
    }

    # Initialize optimization
    # Suppose there are points provided (x0 and y0), record them

    # check x0: list-like, requirement of minimal points
    if x0 is None:
        x0 = []
    elif not isinstance(x0[0], (list, tuple)):
        x0 = [x0]
    if not isinstance(x0, list):
        raise ValueError("`x0` should be a list, but got %s" % type(x0))

    if n_initial_points <= 0 and not x0:
        raise ValueError("Either set `n_initial_points` > 0,"
                         " or provide `x0`")
    # check y0: list-like, requirement of maximal calls
    if isinstance(y0, Iterable):
        y0 = list(y0)
    elif isinstance(y0, numbers.Number):
        y0 = [y0]
    required_calls = n_initial_points + (len(x0) if not y0 else 0)
    if n_calls < required_calls:
        raise ValueError(
            "Expected `n_calls` >= %d, got %d" % (required_calls, n_calls))
    # calculate the total number of initial points
    n_initial_points = n_initial_points + len(x0)

    # Build optimizer

    # create optimizer class
    optimizer = Optimizer(
        dimensions, 
        base_estimator=base_estimator,
        std_estimator=std_estimator,  
        n_initial_points=n_initial_points, 
        initial_point_generator=initial_point_generator,
        acq_func=acq_func, 
        acq_optimizer=acq_optimizer,
        random_state=random_state, 
        acq_func_kwargs=acq_func_kwargs, 
        acq_optimizer_kwargs=acq_optimizer_kwargs,
        base_estimator_kwargs=base_estimator_kwargs,
        std_estimator_kwargs=std_estimator_kwargs,
        model_queue_size=model_queue_size,
        verbose=verbose
    )

    # Record provided points

    # create return object
    result = None
    # evaluate y0 if only x0 is provided
    if x0 and y0 is None:
        y0 = list(map(func, x0))
        n_calls -= len(y0)
    # record through tell function
    if x0:
        if not (isinstance(y0, Iterable) or isinstance(y0, numbers.Number)):
            raise ValueError(
                "`y0` should be an iterable or a scalar, got %s" % type(y0))
        if len(x0) != len(y0):
            raise ValueError("`x0` and `y0` should have the same length")
        result = optimizer.tell(x0, y0)
        result.specs = specs
        if eval_callbacks(callbacks, result):
            return result

    # Handle solver output
    if not isinstance(verbose, (int, type(None))):
        raise TypeError("verbose should be an int of [0,1,2] or bool, "
                        "got {}".format(type(verbose)))

    if isinstance(verbose, bool):
        if verbose:
            verbose = 1
        else:
            verbose = 0
    elif isinstance(verbose, int):
        if verbose not in [0,1,2]:
            raise TypeError("if verbose is int, it should in [0,1,2], "
                        "got {}".format(verbose))

    # Optimize
    _n_calls = n_calls

    itr = 1

    if verbose > 0:
        print("")
        print("SOLVER: start solution process...")
        print("")
        print(f"SOLVER: generate \033[1m {n_initial_points}\033[0m initial points...")


    while _n_calls>0:

        # check if optimization is performed in batches
        if batch_size is not None:
            _batch_size = min([_n_calls,batch_size])
            _n_calls -= _batch_size
            next_x = optimizer.ask(_batch_size,strategy=batch_strategy)
        else:
            _n_calls -= 1
            next_x = optimizer.ask(strategy=batch_strategy)
            
        next_y = func(next_x)

        # first iteration uses next_y as best point instead of min of next_y
        if itr == 1:
            if batch_size is None:
                best_fun = next_y
            else:
                best_fun = min(next_y)

        # handle output print at every iteration
        if verbose > 0:
            print("")
            print(f"\033[1m itr_{itr}\033[0m")

            if isinstance(next_y,Iterable):
                # in case of batch optimization, print all new proposals and 
                # mark improvements of objectives with (*)
                print_str = []
                for y in next_y:
                    if y <= best_fun:
                        print_str.append(f"\033[1m{round(y,5)}\033[0m (*)")
                    else:
                        print_str.append(str(round(y,5)))
                print(f"   new points obj.: {print_str[0]}")
                for y_str in print_str[1:]:
                    print(f"                    {y_str}")    
            else:
                # in case of single point sequential optimization, print new 
                # point proposal
                if next_y <= best_fun:
                    print_str = f"\033[1m itr_{round(next_y,5)}\033[0m"
                else:
                    print_str = str(round(next_y,5))
                print(f"   new point obj.: {round(next_y,5)}")

            # print best obj until (not including) current iteration
            print(f"   best obj.:       {round(best_fun,5)}")

        itr += 1

        res = optimizer.tell(
            next_x, next_y,
            fit= batch_size is None and not _n_calls <=0
        )

        best_fun = res.fun  

        res.specs = specs
    
    # print end of solve once convergence criteria is met
    if verbose>0:
        print("")
        print("SOLVER: finished run!")
        print(f"SOLVER: best obj.: {round(best_fun,5)}")
        print("")

    return res